# -*- coding: utf-8 -*-
"""Sweep config interface."""
from .cfg import SweepConfig, schema_violations_from_proposed_config
from .schema import fill_validate_schema, fill_parameter, fill_validate_early_terminate

__all__ = [
    "SweepConfig",
    "schema_violations_from_proposed_config",
    "fill_validate_schema",
    "fill_parameter",
    "fill_validate_early_terminate",
]



import pytest

from typing import List, Sequence, Tuple
from ..run import RunState, SweepRun, next_run
from ..config import SweepConfig


def kernel_for_grid_search_tests(
    runs: List[SweepRun],
    config: SweepConfig,
    answers: Sequence[Tuple],
    randomize: bool,
) -> None:
    """This kernel assumes that sweep config has two categorical parameters
    named v1 and v2."""

    suggested_parameters = [
        (
            run.config["v1"]["value"],
            run.config["v2"]["value"],
        )
        for run in runs
    ]

    while True:
        suggestion = next_run(config, runs, randomize_order=randomize)
        if suggestion is None:  # done
            break
        assert suggestion.search_info is None
        assert suggestion.state == RunState.pending
        runs.append(suggestion)
        suggested_parameters.append(
            (
                suggestion.config["v1"]["value"],
                suggestion.config["v2"]["value"],
            )
        )

    assert len(answers) == len(suggested_parameters)
    for key in suggested_parameters:
        assert key in answers


@pytest.mark.parametrize("randomize", [True, False])
def test_grid_from_start_with_and_without_randomize(
    sweep_config_2params_grid_search, randomize
):
    kernel_for_grid_search_tests(
        [],
        sweep_config_2params_grid_search,
        randomize=randomize,
        answers=[(1, 4), (1, 5), (2, 4), (2, 5), (3, 4), (3, 5)],
    )


@pytest.mark.parametrize("randomize", [True, False])
def test_grid_search_starting_from_in_progress(
    sweep_config_2params_grid_search, randomize
):
    runs = [
        SweepRun(config={"v1": {"value": 2}, "v2": {"value": 4}}),
        SweepRun(config={"v1": {"value": 1}, "v2": {"value": 5}}),
    ]
    kernel_for_grid_search_tests(
        runs,
        sweep_config_2params_grid_search,
        randomize=randomize,
        answers=[(1, 4), (1, 5), (2, 4), (2, 5), (3, 4), (3, 5)],
    )


def test_grid_search_with_list_values():
    # https://sentry.io/organizations/weights-biases/issues/2501125152/?project=5812400&query=is%3Aresolved&statsPeriod=14d
    config = SweepConfig(
        {
            "method": "grid",
            "parameters": {
                "v1": {
                    "values": ["", [9, 5]],
                },
                "v2": {
                    "values": [256, 512],
                },
            },
        }
    )

    kernel_for_grid_search_tests(
        [],
        config,
        randomize=False,
        answers=[("", 256), ("", 512), ([9, 5], 256), ([9, 5], 512)],
    )


def test_grid_search_duplicated_values_are_not_duplicated_in_answer():
    duplicated_config = SweepConfig(
        {
            "method": "grid",
            "parameters": {
                "v1": {"values": [None, 2, 3, "a", (2, 3), 3]},
                "v2": {"values": ["a", "b", "c'"]},
            },
        }
    )

    runs = []
    kernel_for_grid_search_tests(
        runs,
        duplicated_config,
        randomize=True,
        answers=[
            (
                None,
                "a",
            ),
            (
                2,
                "a",
            ),
            (
                3,
                "a",
            ),
            ("a", "a"),
            (
                (2, 3),
                "a",
            ),
            (
                None,
                "b",
            ),
            (
                2,
                "b",
            ),
            (
                3,
                "b",
            ),
            ((2, 3), "b"),
            ("a", "b"),
            (
                None,
                "c'",
            ),
            (
                2,
                "c'",
            ),
            (
                3,
                "c'",
            ),
            (
                (2, 3),
                "c'",
            ),
            ("a", "c'"),
        ],
    )
    assert len(runs) == 15


def test_grid_search_constant_val_is_propagated():
    config_const = SweepConfig(
        {
            "method": "grid",
            "parameters": {
                "v1": {"values": ["a", "b", "c'"]},
                "v2": {"value": 1},
            },
        }
    )

    runs = []
    run = next_run(config_const, runs)
    assert "v1" in run.config
    assert "v2" in run.config


@pytest.mark.parametrize("randomize", [True, False])
def test_grid_search_dict_val_is_propagated(randomize):
    config_const = SweepConfig(
        {
            "method": "grid",
            "parameters": {
                "v1": {"values": ["a", "b", "c'"]},
                "v2": {
                    "values": [
                        {"a": "b"},
                        {"c": "d", "b": "g"},
                        {"e": {"f": "g"}},
                        {"a": "b"},
                        {"b": "g", "c": "d"},
                    ]
                },
            },
        }
    )

    kernel_for_grid_search_tests(
        [],
        config_const,
        randomize=randomize,
        answers=[
            ("a", {"a": "b"}),
            ("a", {"c": "d", "b": "g"}),
            ("a", {"e": {"f": "g"}}),
            ("b", {"a": "b"}),
            ("b", {"c": "d", "b": "g"}),
            ("b", {"e": {"f": "g"}}),
            ("c'", {"a": "b"}),
            ("c'", {"c": "d", "b": "g"}),
            ("c'", {"e": {"f": "g"}}),
        ],
    )



import itertools
import random
from typing import List, Optional, Union

from .config.cfg import SweepConfig
from .run import SweepRun
from .params import HyperParameter, HyperParameterSet


def grid_search_next_run(
    runs: List[SweepRun],
    sweep_config: Union[dict, SweepConfig],
    validate: bool = False,
    randomize_order: bool = False,
) -> Optional[SweepRun]:
    """Suggest runs with Hyperparameters drawn from a grid.

    >>> suggestion = grid_search_next_run([], {'method': 'grid', 'parameters': {'a': {'values': [1, 2, 3]}}})
    >>> assert suggestion.config['a']['value'] == 1

    Args:
        runs: The runs in the sweep.
        sweep_config: The sweep's config.
        randomize_order: Whether to randomize the order of the grid search.
        validate: Whether to validate `sweep_config` against the SweepConfig JSONschema.
           If true, will raise a Validation error if `sweep_config` does not conform to
           the schema. If false, will attempt to run the sweep with an unvalidated schema.

    Returns:
        The suggested run.
    """

    # make sure the sweep config is valid
    if validate:
        sweep_config = SweepConfig(sweep_config)

    if sweep_config["method"] != "grid":
        raise ValueError("Invalid sweep configuration for grid_search_next_run.")

    if "parameters" not in sweep_config:
        raise ValueError('Grid search requires "parameters" section')
    params = HyperParameterSet.from_config(sweep_config["parameters"])

    # Check that all parameters are categorical or constant
    for p in params:
        if p.type != HyperParameter.CATEGORICAL and p.type != HyperParameter.CONSTANT:
            raise ValueError(
                "Parameter %s is a disallowed type with grid search. Grid search requires all parameters to be categorical or constant"
                % p.name
            )

    # we can only deal with discrete params in a grid search
    discrete_params = [p for p in params if p.type == HyperParameter.CATEGORICAL]

    # build an iterator over all combinations of param values
    param_names = [p.name for p in discrete_params]
    param_values = [p.config["values"] for p in discrete_params]
    param_value_set = list(itertools.product(*param_values))

    if randomize_order:
        random.shuffle(param_value_set)

    new_value_set = next(
        (
            value_set
            for value_set in param_value_set
            # check if parameter set is contained in some run
            if not _runs_contains_param_values(runs, dict(zip(param_names, value_set)))
        ),
        None,
    )

    # handle the case where we couldn't find a unique parameter set
    if new_value_set is None:
        return None

    # set next_run_params based on our new set of params
    for param, value in zip(discrete_params, new_value_set):
        param.value = value

    return SweepRun(config=params.to_config())


def _run_contains_param_values(run, params):
    for key, value in params.items():
        if key not in run.config:
            return False
        if not run.config[key]["value"] == value:
            return False
    return True


def _runs_contains_param_values(runs, params):
    return any(_run_contains_param_values(run, params) for run in runs)



import pytest
import jsonschema
from ..params import HyperParameter
from ..config import SweepConfig, fill_validate_early_terminate
from ..hyperband_stopping import hyperband_baseline_validate_and_fill


def test_json_type_inference_int_uniform():
    config = {"min": 0, "max": 1}
    param = HyperParameter("int_unif_param", config)
    assert param.type == HyperParameter.INT_UNIFORM

    # len is 3 because distribution key is inferred via default
    assert len(param.config) == 3


def test_json_type_inference_uniform():
    config = {"min": 0.0, "max": 1.0}
    param = HyperParameter("unif_param", config)
    assert param.type == HyperParameter.UNIFORM

    # len is 3 because distribution key is inferred via default
    assert len(param.config) == 3


def test_json_type_inference_uniform_mixed():
    config = {"min": 0.0, "max": 1}
    param = HyperParameter("unif_param", config)
    assert param.type == HyperParameter.UNIFORM

    # len is 3 because distribution key is inferred via default
    assert len(param.config) == 3

    config = {"min": 0, "max": 1.0}
    param = HyperParameter("unif_param", config)
    assert param.type == HyperParameter.UNIFORM

    # len is 3 because distribution key is inferred via default
    assert len(param.config) == 3


def test_json_type_inference_and_imputation_normal():
    config = {"distribution": "normal"}
    param = HyperParameter("normal_param", config)
    assert param.config["mu"] == 0
    assert param.config["sigma"] == 1
    assert param.type == HyperParameter.NORMAL
    assert len(param.config) == 3


def test_json_type_inference_and_imputation_lognormal():
    config = {"distribution": "log_normal"}
    param = HyperParameter("log_normal_param", config)
    assert param.config["mu"] == 0
    assert param.config["sigma"] == 1
    assert param.type == HyperParameter.LOG_NORMAL
    assert len(param.config) == 3


def test_json_type_inference_categorical():
    config = {"values": [1, 2, 3]}
    param = HyperParameter("categorical_param", config)
    assert param.type == HyperParameter.CATEGORICAL
    # len is 2 because distribution key is inferred via default
    assert len(param.config) == 2


def test_json_type_inference_constant():
    config = {"value": "abcd"}
    param = HyperParameter("constant_param", config)
    assert param.type == HyperParameter.CONSTANT

    # len is 2 because distribution key is inferred via default
    assert len(param.config) == 2


def test_json_type_inference_q_normal():
    config = {"distribution": "q_normal", "q": 0.1}
    param = HyperParameter("q_normal_param", config)
    assert param.config["mu"] == 0
    assert param.config["sigma"] == 1
    assert param.type == HyperParameter.Q_NORMAL
    assert len(param.config) == 4


def test_json_type_inference_q_beta():
    config = {"distribution": "q_beta"}
    param = HyperParameter("q_beta_param", config)
    assert param.config["a"] == 1
    assert param.config["b"] == 1
    assert param.config["q"] == 1
    assert param.type == HyperParameter.Q_BETA
    assert len(param.config) == 4


def test_json_type_inference_beta():
    config = {"distribution": "beta"}
    param = HyperParameter("beta_param", config)
    assert param.config["a"] == 1
    assert param.config["b"] == 1
    assert param.type == HyperParameter.BETA
    assert len(param.config) == 3


def test_totally_invalid_config():
    config = {"invalid": "this config is not valid"}
    with pytest.raises(jsonschema.ValidationError):
        HyperParameter("invalid", config)


def test_validate_does_not_modify_passed_config():
    config = {"distribution": "normal"}
    config_save = config.copy()
    _ = HyperParameter("normal_test", config)
    assert config == config_save


def test_categorical_hyperparameter_no_values():
    config = {"values": []}
    with pytest.raises(jsonschema.ValidationError):
        HyperParameter("invalid_test", config)


def test_uniform_with_integer_min_max():
    # CLI-975
    # https://github.com/wandb/sweeps/pull/16
    config = {"distribution": "uniform", "min": 0, "max": 1}  # integers
    unif_param = HyperParameter("uniform_param", config)  # this should not raise
    assert unif_param.type == HyperParameter.UNIFORM


def test_hyperband_missing_eta_imputed():
    # sentry https://sentry.io/organizations/weights-biases/issues/2500192925/?referrer=slack
    config = {
        "command": [
            "${env}",
            "${interpreter}",
            "${program}",
            "sweep",
            "2dconv_lstm.stat.ac",
            "--dataset",
            "deam",
            "--temp_folder",
            "a_deam",
            "--batch-size",
        ],
        "early_terminate": {"min_iter": 3, "type": "hyperband"},
        "method": "random",
        "metric": {"goal": "minimize", "name": "val/loss"},
        "name": "AC-2DConvLSTM-Stat-DEAM",
        "parameters": {
            "dropout": {"values": ["0.15", "0.2", "0.25", "0.3", "0.4", "0.5"]},
            "lr": {"values": ["0.001", "0.005", "0.01"]},
            "momentum": {"values": ["0.8", "0.9", "0.95"]},
            "n_fft": {"value": 1024},
            "n_mels": {"value": 128},
        },
        "program": "exec.py",
        "project": "mer",
    }

    sc = SweepConfig(config)
    assert sc["early_terminate"]["eta"] == 3


def test_hyperband_missing_eta_imputed_incremental():
    # sentry https://sentry.io/organizations/weights-biases/issues/2500192925/?referrer=slack
    config = {
        "command": [
            "${env}",
            "${interpreter}",
            "${program}",
            "sweep",
            "2dconv_lstm.stat.ac",
            "--dataset",
            "deam",
            "--temp_folder",
            "a_deam",
            "--batch-size",
        ],
        "early_terminate": {"min_iter": 3, "type": "hyperband"},
        "method": "random",
        "metric": {"goal": "minimize", "name": "val/loss"},
        "name": "AC-2DConvLSTM-Stat-DEAM",
        "parameters": {
            "dropout": {"values": ["0.15", "0.2", "0.25", "0.3", "0.4", "0.5"]},
            "lr": {"values": ["0.001", "0.005", "0.01"]},
            "momentum": {"values": ["0.8", "0.9", "0.95"]},
            "n_fft": {"value": 1024},
            "n_mels": {"value": 128},
        },
        "program": "exec.py",
        "project": "mer",
    }

    sc = fill_validate_early_terminate(config)
    assert sc["early_terminate"]["eta"] == 3


def test_hyperband_incremental_corrects_goal():
    # sentry https://sentry.io/organizations/weights-biases/issues/2500192925/?referrer=slack
    config = {
        "command": [
            "${env}",
            "${interpreter}",
            "${program}",
            "sweep",
            "2dconv_lstm.stat.ac",
            "--dataset",
            "deam",
            "--temp_folder",
            "a_deam",
            "--batch-size",
        ],
        "early_terminate": {"min_iter": 3, "type": "hyperband"},
        "method": "random",
        "metric": {"goal": "minimise", "name": "val/loss"},
        "name": "AC-2DConvLSTM-Stat-DEAM",
        "parameters": {
            "dropout": {"values": ["0.15", "0.2", "0.25", "0.3", "0.4", "0.5"]},
            "lr": {"values": ["0.001", "0.005", "0.01"]},
            "momentum": {"values": ["0.8", "0.9", "0.95"]},
            "n_fft": {"value": 1024},
            "n_mels": {"value": 128},
        },
        "program": "exec.py",
        "project": "mer",
    }

    sc = hyperband_baseline_validate_and_fill(config)
    assert sc["early_terminate"]["eta"] == 3
    assert sc["metric"]["goal"] == "minimize"



import getpass
import json
import logging
import os
import posixpath
import re
import shutil
import subprocess
import sys
import tempfile
from typing import Any, Dict, List, Optional, Sequence, Union


from docker.models.resource import Model  # type: ignore
from dockerpycreds.utils import find_executable  # type: ignore
from six.moves import shlex_quote
import wandb
from wandb.apis.internal import Api
from wandb.env import DOCKER
from wandb.errors import ExecutionError, LaunchError
from wandb.util import get_module

from . import _project_spec
from .utils import _is_wandb_dev_uri, _is_wandb_local_uri
from ..lib.git import GitRepo

_logger = logging.getLogger(__name__)

_GENERATED_DOCKERFILE_NAME = "Dockerfile.wandb-autogenerated"
_PROJECT_TAR_ARCHIVE_NAME = "wandb-project-docker-build-context"


def validate_docker_installation() -> None:
    """Verify if Docker is installed on host machine."""
    if not find_executable("docker"):
        raise ExecutionError(
            "Could not find Docker executable. "
            "Ensure Docker is installed as per the instructions "
            "at https://docs.docker.com/install/overview/."
        )


def validate_docker_env(launch_project: _project_spec.LaunchProject) -> None:
    """Ensure project has a docker image associated with it."""
    if not launch_project.docker_image:
        raise ExecutionError(
            "LaunchProject with docker environment must specify the docker image "
            "to use via 'docker_image' field."
        )


def generate_docker_image(
    launch_project: _project_spec.LaunchProject, entry_cmd: str
) -> str:
    """Uses project and entry point to generate the docker image."""
    path = launch_project.project_dir
    # this check will always pass since the dir attribute will always be populated
    # by _fetch_project_local
    get_module(
        "repo2docker",
        required='wandb launch requires additional dependencies, install with pip install "wandb[launch]"',
    )
    assert isinstance(path, str)
    cmd: Sequence[str] = [
        "jupyter-repo2docker",
        "--no-run",
        "--user-id={}".format(launch_project.docker_user_id),
        path,
        '"{}"'.format(entry_cmd),
    ]

    _logger.info(
        "Generating docker image from git repo or finding image if it already exists.........."
    )
    wandb.termlog("Generating docker image, this may take a few minutes")
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stderr = ""
    # this will always pass, repo2docker writes to stderr.
    assert process.stderr
    for line in process.stderr:
        decoded_line = line.decode("utf-8")
        if decoded_line.endswith("\n"):
            decoded_line = decoded_line.rstrip("\n")
        print(decoded_line)  # don't spam termlog with all this
        stderr = stderr + decoded_line
    process.wait()
    image_id: List[str] = re.findall(r"Successfully tagged (.+):latest", stderr)
    if not image_id:
        image_id = re.findall(r"Reusing existing image \((.+)\)", stderr)
    if not image_id:
        raise LaunchError("error running repo2docker: {}".format(stderr))
    os.environ[DOCKER] = image_id[0]
    return image_id[0]


def pull_docker_image(docker_image: str) -> None:
    """Pulls the requested docker image."""
    import docker  # type: ignore

    info = docker_image.split(":")
    client = docker.from_env()
    try:
        if len(info) == 1:
            client.images.pull(info[0])
        else:
            client.images.pull(info[0], tag=info[1])
    except docker.errors.APIError as e:
        raise LaunchError("Docker server returned error: {}".format(e))


def build_docker_image(
    launch_project: _project_spec.LaunchProject, base_image: str, copy_code: bool,
) -> Union[Model, Any]:
    """Build a docker image containing the project in `work_dir`, using the base image.

    Arguments:
    launch_project: LaunchProject class instance
    base_image: base_image to build the docker image off of
    api: instance of wandb.apis.internal Api
    copy_code: boolean indicating if code should be copied into the docker container

    Returns:
        A `Model` instance of the docker image.

    Raises:
        LaunchError: if there is an issue communicating with the docker client
    """
    import docker

    image_name = "wandb_launch_{}".format(launch_project.run_id)
    image_uri = _get_docker_image_uri(
        name=image_name, work_dir=launch_project.project_dir
    )
    copy_code_line = ""
    workdir_line = ""
    copy_config_line = ""
    workdir = os.path.join("/home/", getpass.getuser())
    if launch_project.override_config:
        copy_config_line = "COPY {}/{} {}\n".format(
            _PROJECT_TAR_ARCHIVE_NAME, _project_spec.DEFAULT_CONFIG_PATH, workdir
        )
    if copy_code:
        copy_code_line = "COPY {}/ {}\n".format(_PROJECT_TAR_ARCHIVE_NAME, workdir)
        workdir_line = "WORKDIR {}\n".format(workdir)
    name_line = ""
    if launch_project.name:
        name_line = "ENV WANDB_NAME={wandb_name}\n"

    dockerfile = (
        "FROM {imagename}\n"
        "{copy_config_line}"
        "{copy_code_line}"
        "{workdir_line}"
        "{name_line}"
    ).format(
        imagename=base_image,
        copy_config_line=copy_config_line,
        copy_code_line=copy_code_line,
        workdir_line=workdir_line,
        name_line=name_line,
    )
    build_ctx_path = _create_docker_build_ctx(
        launch_project.project_dir,
        dockerfile,
        launch_project._runtime,
        launch_project.override_config,
    )
    with open(build_ctx_path, "rb") as docker_build_ctx:
        _logger.info("=== Building docker image %s ===", image_uri)
        #  TODO: replace with shelling out
        dockerfile = posixpath.join(
            _PROJECT_TAR_ARCHIVE_NAME, _GENERATED_DOCKERFILE_NAME
        )
        # TODO: remove the dependency on docker / potentially just do the append builder
        # found at: https://github.com/google/containerregistry/blob/master/client/v2_2/append_.py
        client = docker.from_env()
        try:
            image, _ = client.images.build(
                tag=image_uri,
                forcerm=True,
                dockerfile=dockerfile,
                fileobj=docker_build_ctx,
                custom_context=True,
                encoding="gzip",
            )
        except ConnectionError as e:
            raise LaunchError("Error communicating with docker client: {}".format(e))

    try:
        os.remove(build_ctx_path)
    except Exception:
        _logger.info(
            "Temporary docker context file %s was not deleted.", build_ctx_path
        )
    return image


def get_docker_command(
    image: Union[Model, Any],
    launch_project: _project_spec.LaunchProject,
    api: Api,
    docker_args: Dict[str, Any] = None,
) -> List[str]:
    """Constructs the docker command using the image and docker args.

    Arguments:
    image: a Docker image to be run
    launch_project: an instance of LaunchProject
    api: an instance of wandb.apis.internal Api
    docker_args: a dictionary of additional docker args for the command
    """
    docker_path = "docker"
    cmd: List[Any] = [docker_path, "run", "--rm"]

    if _is_wandb_local_uri(api.settings("base_url")) and sys.platform == "darwin":
        _, _, port = _, _, port = api.settings("base_url").split(":")
        base_url = "http://host.docker.internal:{}".format(port)
    elif _is_wandb_dev_uri(api.settings("base_url")):
        base_url = "http://host.docker.internal:9002"
    else:
        base_url = api.settings("base_url")

    cmd += [
        "--env",
        f"WANDB_BASE_URL={base_url}",
        "--env",
        f"WANDB_API_KEY={api.api_key}",
        "--env",
        f"WANDB_PROJECT={launch_project.target_project}",
        "--env",
        f"WANDB_ENTITY={launch_project.target_entity}",
        "--env",
        f"WANDB_LAUNCH={True}",
        "--env",
        f"WANDB_LAUNCH_CONFIG_PATH={_project_spec.DEFAULT_CONFIG_PATH}",
        "--env",
        f"WANDB_RUN_ID={launch_project.run_id or None}",
        "--env",
        f"WANDB_DOCKER={launch_project.docker_image}",
    ]

    if docker_args:
        for name, value in docker_args.items():
            # Passed just the name as boolean flag
            if isinstance(value, bool) and value:
                if len(name) == 1:
                    cmd += ["-" + name]
                else:
                    cmd += ["--" + name]
            else:
                # Passed name=value
                if len(name) == 1:
                    cmd += ["-" + name, value]
                else:
                    cmd += ["--" + name, value]

    cmd += [image.tags[0]]
    return [shlex_quote(c) for c in cmd]


def _get_docker_image_uri(name: str, work_dir: str) -> str:
    """Returns a Docker image URI based on the git hash of the specified working directory.

    Arguments:
    name: The URI of the Docker repository with which to tag the image. The
        repository URI is used as the prefix of the image URI.
    work_dir: Path to the working directory in which to search for a git commit hash
    """
    name = name.replace(" ", "-") if name else "docker-project"
    # Optionally include first 7 digits of git SHA in tag name, if available.

    git_commit = GitRepo(work_dir).last_commit
    version_string = ":" + str(git_commit[:7]) if git_commit else ""
    return name + version_string


def _create_docker_build_ctx(
    work_dir: str,
    dockerfile_contents: str,
    runtime: Optional[str],
    run_config: Dict[str, Any],
) -> str:
    """Creates build context tarfile containing Dockerfile and project code, returning path to tarfile."""
    directory = tempfile.mkdtemp()
    try:
        dst_path = os.path.join(directory, "wandb-project-contents")
        shutil.copytree(src=work_dir, dst=dst_path)
        if run_config:
            config_path = os.path.join(dst_path, _project_spec.DEFAULT_CONFIG_PATH)
            with open(config_path, "w") as fp:
                json.dump(run_config, fp)
        if runtime:
            runtime_path = os.path.join(dst_path, "runtime.txt")
            with open(runtime_path, "w") as fp:
                fp.write(runtime)

        with open(os.path.join(dst_path, _GENERATED_DOCKERFILE_NAME), "w") as handle:
            handle.write(dockerfile_contents)
        _, result_path = tempfile.mkstemp()
        wandb.util.make_tarfile(
            output_filename=result_path,
            source_dir=dst_path,
            archive_name=_PROJECT_TAR_ARCHIVE_NAME,
        )
    finally:
        shutil.rmtree(directory)
    return result_path



import pytest
from jsonschema import ValidationError
from .. import next_run, stop_runs, SweepRun
from ..config import SweepConfig, schema_violations_from_proposed_config
from ..bayes_search import bayes_search_next_run
from ..grid_search import grid_search_next_run
from ..random_search import random_search_next_run
from ..hyperband_stopping import hyperband_stop_runs


@pytest.mark.parametrize("search_type", ["bayes", "grid", "random"])
def test_validation_disable(search_type):
    invalid_schema = {
        "metric": {"name": "loss", "goal": "minimise"},
        "method": search_type,
        "parameters": {
            "v1": {"values": ["a", "b", "c"]},
        },
    }

    with pytest.raises(ValidationError):
        _ = next_run(invalid_schema, [], validate=True)

    with pytest.raises(ValidationError):
        _ = stop_runs(invalid_schema, [], validate=True)
        _ = hyperband_stop_runs([], invalid_schema, validate=True)

    with pytest.raises(ValidationError):
        if search_type == "bayes":
            _ = bayes_search_next_run([], invalid_schema, validate=True)
        elif search_type == "grid":
            _ = grid_search_next_run([], invalid_schema, validate=True)
        elif search_type == "random":
            _ = random_search_next_run(invalid_schema, validate=True)

    # check that no error is raised
    result = next_run(invalid_schema, [], validate=False)
    assert result is not None


def test_validation_not_enough_params():
    schema = {"method": "random", "parameters": {}}

    with pytest.raises(ValidationError):
        SweepConfig(schema)

    # test that this doesnt raise a keyerror https://sentry.io/organizations/weights-biases/issues/2461042074/?project=5812400&query=is%3Aresolved&statsPeriod=14d
    result = schema_violations_from_proposed_config(schema)
    assert len(result) == 1

    # test that this is rejected at the next_run stage
    with pytest.raises(ValueError):
        next_run(schema, [])

    del schema["parameters"]
    # test that this is rejected at the next_run stage
    with pytest.raises(ValueError):
        next_run(schema, [])


def test_minmax_type_inference():
    schema = {
        "method": "random",
        "parameters": {"a": {"min": 0, "max": 1, "distribution": "uniform"}},
    }

    violations = schema_violations_from_proposed_config(schema)
    assert len(violations) == 0

    schema = {
        "method": "random",
        "parameters": {"a": {"min": 0.0, "max": 1.0, "distribution": "int_uniform"}},
    }

    violations = schema_violations_from_proposed_config(schema)
    assert len(violations) == 1


@pytest.mark.parametrize("controller_type", ["cloud", "local", "invalid"])
def test_controller(controller_type):
    schema = {
        "controller": {"type": controller_type},
        "method": "random",
        "parameters": {"a": {"values": [1, 2, 3, 4]}},
    }

    if controller_type in ["cloud", "local"]:
        assert SweepConfig(schema)["controller"]["type"] == controller_type
    else:
        with pytest.raises(ValidationError):
            SweepConfig(schema)


def test_invalid_config():
    config = "this is a totally invalid config"
    with pytest.raises(ValueError):
        schema_violations_from_proposed_config(config)


def test_sweepconfig_no_method_baseline_validation():
    schema = {
        "parameters": {"a": {"values": [1, 2, 3, 4]}},
    }

    with pytest.raises(ValueError):
        next_run(schema, [], validate=False)


def test_invalid_early_stopping():
    invalid_schema = {
        "method": "bayes",
        "parameters": {
            "v1": {"values": ["a", "b", "c"]},
        },
    }

    with pytest.raises(ValueError):
        stop_runs(invalid_schema, [], validate=False)

    invalid_schema["metric"] = {"name": "loss", "goal": "minimise"}

    with pytest.raises(ValueError):
        stop_runs(invalid_schema, [], validate=False)

    invalid_schema["early_terminate"] = dict()
    invalid_schema["early_terminate"]["type"] = "invalid type"

    with pytest.raises(ValueError):
        stop_runs(invalid_schema, [], validate=False)

    invalid_schema["early_terminate"]["type"] = "hyperband"
    invalid_schema["early_terminate"]["extra_key"] = 1234
    invalid_schema["early_terminate"]["min_iter"] = 100

    to_stop = stop_runs(invalid_schema, [], validate=False)
    assert len(to_stop) == 0


def test_invalid_run_parameter():
    config = {
        "method": "bayes",
        "metric": {"name": "loss", "goal": "minimize"},
        "parameters": {
            "v1": {"values": ["a", "b", "c"]},
        },
    }

    runs = [SweepRun(config={"v1": {"value": "d"}}, summary_metrics={"loss": 5.0})]

    with pytest.raises(ValueError):
        next_run(config, runs, validate=False)



"""Hyperparameter search parameters."""

import random

from typing import List, Tuple, Dict, Any

import numpy as np
import scipy.stats as stats

import jsonschema

from .run import SweepRun
from .config import fill_parameter
from ._types import ArrayLike


class HyperParameter:

    CONSTANT = "param_single_value"
    CATEGORICAL = "param_categorical"
    INT_UNIFORM = "param_int_uniform"
    UNIFORM = "param_uniform"
    LOG_UNIFORM = "param_loguniform"
    Q_UNIFORM = "param_quniform"
    Q_LOG_UNIFORM = "param_qloguniform"
    NORMAL = "param_normal"
    Q_NORMAL = "param_qnormal"
    LOG_NORMAL = "param_lognormal"
    Q_LOG_NORMAL = "param_qlognormal"
    BETA = "param_beta"
    Q_BETA = "param_qbeta"

    def __init__(self, name: str, config: dict):
        """A hyperparameter to optimize.

        >>> parameter = HyperParameter('int_unif_distributed', {'min': 1, 'max': 10})
        >>> assert parameter.config['min'] == 1
        >>> parameter = HyperParameter('normally_distributed', {'distribution': 'normal'})
        >>> assert np.isclose(parameter.config['mu'], 0)

        Args:
            name: The name of the hyperparameter.
            config: Hyperparameter config dict.
        """

        self.name = name

        result = fill_parameter(config)
        if result is None:
            raise jsonschema.ValidationError(
                f"invalid hyperparameter configuration: {name}"
            )

        self.type, self.config = result
        if self.config is None or self.type is None:
            raise ValueError(
                "list of allowed schemas has length zero; please provide some valid schemas"
            )

        self.value = (
            None if self.type != HyperParameter.CONSTANT else self.config["value"]
        )

    def value_to_int(self, value: Any) -> int:
        """Get the index of the value of a categorically distributed HyperParameter.

        >>> parameter = HyperParameter('a', {'values': [1, 2, 3]})
        >>> assert parameter.value_to_int(2) == 1

        Args:
             value: The value to look up.

        Returns:
            The index of the value.
        """

        if self.type != HyperParameter.CATEGORICAL:
            raise ValueError("Can only call value_to_int on categorical variable")

        for ii, test_value in enumerate(self.config["values"]):
            if value == test_value:
                return ii

        raise ValueError(
            f"{value} is not a permitted value of the categorical hyperparameter {self.name} "
            f"in the current sweep."
        )

    def cdf(self, x: ArrayLike) -> ArrayLike:
        """Cumulative distribution function (CDF).

        In probability theory and statistics, the cumulative distribution function
        (CDF) of a real-valued random variable X, is the probability that X will
        take a value less than or equal to x.

        Args:
             x: Parameter values to calculate the CDF for. Can be scalar or 1-d.
        Returns:
            Probability that a random sample of this hyperparameter will be less than x.
        """
        if self.type == HyperParameter.CONSTANT:
            return np.zeros_like(x)
        elif self.type == HyperParameter.CATEGORICAL:
            # NOTE: Indices expected for categorical parameters, not values.
            return stats.randint.cdf(x, 0, len(self.config["values"]))
        elif self.type == HyperParameter.INT_UNIFORM:
            return stats.randint.cdf(x, self.config["min"], self.config["max"] + 1)
        elif (
            self.type == HyperParameter.UNIFORM or self.type == HyperParameter.Q_UNIFORM
        ):
            return stats.uniform.cdf(
                x, self.config["min"], self.config["max"] - self.config["min"]
            )
        elif (
            self.type == HyperParameter.LOG_UNIFORM
            or self.type == HyperParameter.Q_LOG_UNIFORM
        ):
            return stats.uniform.cdf(
                np.log(x), self.config["min"], self.config["max"] - self.config["min"]
            )
        elif self.type == HyperParameter.NORMAL or self.type == HyperParameter.Q_NORMAL:
            return stats.norm.cdf(x, loc=self.config["mu"], scale=self.config["sigma"])
        elif (
            self.type == HyperParameter.LOG_NORMAL
            or self.type == HyperParameter.Q_LOG_NORMAL
        ):
            return stats.lognorm.cdf(
                x, s=self.config["sigma"], scale=np.exp(self.config["mu"])
            )
        elif self.type == HyperParameter.BETA or self.type == HyperParameter.Q_BETA:
            return stats.beta.cdf(x, a=self.config["a"], b=self.config["b"])
        else:
            raise ValueError("Unsupported hyperparameter distribution type")

    def ppf(self, x: ArrayLike) -> Any:
        """Percentage point function (PPF).

        In probability theory and statistics, the percentage point function is
        the inverse of the CDF: it returns the value of a random variable at the
        xth percentile.

        Args:
             x: Percentiles of the random variable. Can be scalar or 1-d.
        Returns:
            Value of the random variable at the specified percentile.
        """
        if np.any((x < 0.0) | (x > 1.0)):
            raise ValueError("Can't call ppf on value outside of [0,1]")
        if self.type == HyperParameter.CONSTANT:
            return self.config["value"]
        elif self.type == HyperParameter.CATEGORICAL:
            retval = [
                self.config["values"][i]
                for i in np.atleast_1d(
                    stats.randint.ppf(x, 0, len(self.config["values"])).astype(int)
                ).tolist()
            ]
            if np.isscalar(x):
                return retval[0]
            return retval
        elif self.type == HyperParameter.INT_UNIFORM:
            return (
                stats.randint.ppf(x, self.config["min"], self.config["max"] + 1)
                .astype(int)
                .tolist()
            )
        elif self.type == HyperParameter.UNIFORM:
            return stats.uniform.ppf(
                x, self.config["min"], self.config["max"] - self.config["min"]
            )
        elif self.type == HyperParameter.Q_UNIFORM:
            r = stats.uniform.ppf(
                x, self.config["min"], self.config["max"] - self.config["min"]
            )
            ret_val = np.round(r / self.config["q"]) * self.config["q"]
            if isinstance(self.config["q"], int):
                return ret_val.astype(int)
            else:
                return ret_val
        elif self.type == HyperParameter.LOG_UNIFORM:
            return np.exp(
                stats.uniform.ppf(
                    x, self.config["min"], self.config["max"] - self.config["min"]
                )
            )
        elif self.type == HyperParameter.Q_LOG_UNIFORM:
            r = np.exp(
                stats.uniform.ppf(
                    x, self.config["min"], self.config["max"] - self.config["min"]
                )
            )
            ret_val = np.round(r / self.config["q"]) * self.config["q"]
            if isinstance(self.config["q"], int):
                return ret_val.astype(int)
            else:
                return ret_val
        elif self.type == HyperParameter.NORMAL:
            return stats.norm.ppf(x, loc=self.config["mu"], scale=self.config["sigma"])
        elif self.type == HyperParameter.Q_NORMAL:
            r = stats.norm.ppf(x, loc=self.config["mu"], scale=self.config["sigma"])
            ret_val = np.round(r / self.config["q"]) * self.config["q"]
            if isinstance(self.config["q"], int):
                return ret_val.astype(int)
            else:
                return ret_val
        elif self.type == HyperParameter.LOG_NORMAL:
            # https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.lognorm.html
            return stats.lognorm.ppf(
                x, s=self.config["sigma"], scale=np.exp(self.config["mu"])
            )
        elif self.type == HyperParameter.Q_LOG_NORMAL:
            r = stats.lognorm.ppf(
                x, s=self.config["sigma"], scale=np.exp(self.config["mu"])
            )
            ret_val = np.round(r / self.config["q"]) * self.config["q"]

            if isinstance(self.config["q"], int):
                return ret_val.astype(int)
            else:
                return ret_val

        elif self.type == HyperParameter.BETA:
            return stats.beta.ppf(x, a=self.config["a"], b=self.config["b"])
        elif self.type == HyperParameter.Q_BETA:
            r = stats.beta.ppf(x, a=self.config["a"], b=self.config["b"])
            ret_val = np.round(r / self.config["q"]) * self.config["q"]
            if isinstance(self.config["q"], int):
                return ret_val.astype(int)
            else:
                return ret_val
        else:
            raise ValueError("Unsupported hyperparameter distribution type")

    def sample(self) -> Any:
        """Randomly sample a value from the distribution of this HyperParameter."""
        return self.ppf(random.uniform(0.0, 1.0))

    def _to_config(self) -> Tuple[str, Dict]:
        config = dict(value=self.value)
        return self.name, config


class HyperParameterSet(list):
    def __init__(self, items: List[HyperParameter]):
        """A set of HyperParameters.

        >>> hp1 = HyperParameter('a', {'values': [1, 2, 3]})
        >>> hp2 = HyperParameter('b', {'distribution': 'normal'})
        >>> HyperParameterSet([hp1, hp2])

        Args:
            items: A list of HyperParameters to construct the set from.
        """

        for item in items:
            if not isinstance(item, HyperParameter):
                raise TypeError(
                    f"each item used to initialize HyperParameterSet must be a HyperParameter, got {item}"
                )

        super().__init__(items)
        self.searchable_params = [
            param for param in self if param.type != HyperParameter.CONSTANT
        ]

        self.param_names_to_index = {}
        self.param_names_to_param = {}

        for ii, param in enumerate(self.searchable_params):
            self.param_names_to_index[param.name] = ii
            self.param_names_to_param[param.name] = param

    @classmethod
    def from_config(cls, config: Dict):
        """Instantiate a HyperParameterSet based the parameters section of a SweepConfig.

        >>> sweep_config = {'method': 'grid', 'parameters': {'a': {'values': [1, 2, 3]}}}
        >>> hps = HyperParameterSet.from_config(sweep_config['parameters'])

        Args:
            config: The parameters section of a SweepConfig.
        """
        hpd = cls(
            [
                HyperParameter(param_name, param_config)
                for param_name, param_config in sorted(config.items())
            ]
        )
        return hpd

    def to_config(self) -> Dict:
        """Convert a HyperParameterSet to a SweepRun config."""
        return dict([param._to_config() for param in self])

    def convert_runs_to_normalized_vector(self, runs: List[SweepRun]) -> ArrayLike:
        """Converts a list of SweepRuns to an array of normalized parameter vectors.

        Args:
            runs: List of runs to convert.

        Returns:
            A 2d array of normalized parameter vectors.
        """

        runs_params = [run.config for run in runs]
        X = np.zeros([len(self.searchable_params), len(runs)])

        for key, bayes_opt_index in self.param_names_to_index.items():
            param = self.param_names_to_param[key]
            row = np.array(
                [
                    (
                        param.value_to_int(config[key]["value"])
                        if param.type == HyperParameter.CATEGORICAL
                        else config[key]["value"]
                    )
                    if key in config
                    # filter out incorrectly specified runs
                    else np.nan
                    for config in runs_params
                ]
            )

            X_row = param.cdf(row)
            # only use values where input wasn't nan
            non_nan = ~np.isnan(row)
            X[bayes_opt_index, non_nan] = X_row[non_nan]

        return np.transpose(X)



#
import logging
import sys

import wandb

logger = logging.getLogger(__name__)


STYLED_TABLE_HTML = """<style>
    table.wandb td:nth-child(1) { padding: 0 10px; text-align: right }
    </style><table class="wandb">
"""


def _get_python_type():
    try:
        from IPython import get_ipython  # type: ignore

        # Calling get_ipython can cause an ImportError
        if get_ipython() is None:
            return "python"
    except ImportError:
        return "python"
    if "terminal" in get_ipython().__module__ or "spyder" in sys.modules:
        return "ipython"
    else:
        return "jupyter"


def display_html(html):
    """Displays HTML in notebooks, is a noop outside of a jupyter context"""
    if wandb.run and wandb.run._settings._silent:
        return
    try:
        from IPython.core.display import display, HTML  # type: ignore
    except ImportError:
        wandb.termwarn("Unable to render HTML, can't import display from ipython.core")
        return False
    return display(HTML(html))


def display_widget(widget):
    """Displays ipywidgets in notebooks, is a noop outside of a jupyter context"""
    if wandb.run and wandb.run._settings._silent:
        return
    try:
        from IPython.core.display import display
    except ImportError:
        wandb.termwarn(
            "Unable to render Widget, can't import display from ipython.core"
        )
        return False
    return display(widget)


class ProgressWidget(object):
    """A simple wrapper to render a nice progress bar with a label"""

    def __init__(self, widgets, min, max):
        self.widgets = widgets
        self._progress = widgets.FloatProgress(min=min, max=max)
        self._label = widgets.Label()
        self._widget = self.widgets.VBox([self._label, self._progress])
        self._displayed = False
        self._disabled = False

    def update(self, value, label):
        if self._disabled:
            return
        try:
            self._progress.value = value
            self._label.value = label
            if not self._displayed:
                self._displayed = True
                display_widget(self._widget)
        except Exception as e:
            self._disabled = True
            logger.exception(e)
            wandb.termwarn(
                "Unable to render progress bar, see the user log for details"
            )

    def close(self):
        if self._disabled or not self._displayed:
            return
        self._widget.close()


def jupyter_progress_bar(min=0, max=1.0):
    """Returns an ipywidget progress bar or None if we can't import it"""
    widgets = wandb.util.get_module("ipywidgets")
    try:
        if widgets is None:
            # TODO: this currently works in iPython but it's deprecated since 4.0
            from IPython.html import widgets  # type: ignore

        assert hasattr(widgets, "VBox")
        assert hasattr(widgets, "Label")
        assert hasattr(widgets, "FloatProgress")
        return ProgressWidget(widgets, min=min, max=max)
    except (ImportError, AssertionError):
        return None



import os
import json
from typing import Callable, Optional, Tuple, Iterable, Dict, Union

import pytest
import numpy as np

from .. import bayes_search as bayes
from .._types import integer, floating, ArrayLike
from .. import SweepRun, RunState, next_run, SweepConfig

from .test_random_search import check_that_samples_are_from_the_same_distribution


def squiggle(x: ArrayLike) -> np.floating:
    # the maximum of this 1d function is at x=2 and the minimum is at ~3.6 over the
    # interval 0-5
    return np.exp(-((x - 2) ** 2)) + np.exp(-((x - 6) ** 2) / 10) + 1 / (x ** 2 + 1)


def rosenbrock(x: ArrayLike) -> np.floating:
    # has a minimum at (1, 1, 1, 1, ...) for 4 <= ndim <= 7
    return np.sum((x[1:] - x[:-1] ** 2.0) ** 2.0 + (1 - x[:-1]) ** 2.0)


def run_bayes_search(
    f: Callable[[SweepRun], floating],
    config: SweepConfig,
    init_runs: Iterable[SweepRun] = (),
    improvement: floating = 0.1,
    num_iterations: integer = 20,
    optimium: Optional[Dict[str, floating]] = None,
    atol: Optional[Union[Dict[str, floating], floating]] = 0.2,
):

    metric_name = config["metric"]["name"]
    opt_goal = config["metric"]["goal"]

    runs = list(init_runs)
    for _ in range(num_iterations):
        suggested_run = bayes.bayes_search_next_run(
            runs, config, minimum_improvement=improvement
        )
        suggested_run.state = RunState.finished
        metric = f(suggested_run)
        if suggested_run.summary_metrics is None:
            suggested_run.summary_metrics = {}  # pragma: no cover
        suggested_run.summary_metrics[metric_name] = metric
        runs.append(suggested_run)

    if optimium is not None:
        best_run = (min if opt_goal == "minimize" else max)(
            runs, key=lambda run: run.metric_extremum(metric_name, opt_goal)
        )
        for param_name in config["parameters"]:
            left_comp = best_run.config[param_name]["value"]
            right_comp = (
                optimium if isinstance(optimium, float) else optimium[param_name]
            )
            np.testing.assert_allclose(left_comp, right_comp, atol=atol)


@pytest.mark.parametrize(
    "x",
    [
        {"distribution": "normal", "mu": 2, "sigma": 4},
        {"distribution": "log_uniform", "min": -2, "max": 3},
        {"min": 0.0, "max": 5.0},
        {"min": 0, "max": 5},
        {"distribution": "q_uniform", "min": 0.0, "max": 10.0, "q": 0.25},
    ],
)
def test_squiggle_convergence_full(x):
    def y(x: SweepRun) -> floating:
        return squiggle(x.config["x"]["value"])

    run = SweepRun(
        config={"x": {"value": np.random.uniform(0, 5)}}, state=RunState.finished
    )
    run.summary_metrics["y"] = y(run)

    runs = [run]

    config = SweepConfig(
        {
            "method": "bayes",
            "metric": {"name": "y", "goal": "maximize"},
            "parameters": {"x": x},
        }
    )

    run_bayes_search(y, config, runs, num_iterations=200, optimium={"x": 2.0})


def run_iterations(
    f: Callable[[ArrayLike], floating],
    bounds: ArrayLike,
    num_iterations: integer = 20,
    x_init: Optional[ArrayLike] = None,
    improvement: floating = 0.1,
    optimium: Optional[ArrayLike] = None,
    atol: Optional[ArrayLike] = 0.2,
    chunk_size: integer = 1,
) -> Tuple[ArrayLike, ArrayLike]:

    if x_init is not None:
        X = x_init
    else:
        X = [np.zeros(len(bounds))]

    y = np.array([f(x) for x in X]).flatten()

    counter = 0
    for jj in range(int(np.ceil(num_iterations / chunk_size))):
        sample_X = None
        for cc in range(chunk_size):
            if counter >= num_iterations:
                break
            (sample, prob, pred, _, _,) = bayes.next_sample(
                sample_X=X,
                sample_y=y,
                X_bounds=bounds,
                current_X=sample_X,
                improvement=improvement,
            )
            if sample_X is None:
                sample_X = np.array([sample])
            else:
                sample_X = np.append(sample_X, np.array([sample]), axis=0)
            counter += 1
            print(
                "X: {} prob(I): {} pred: {} value: {}".format(
                    sample, prob, pred, f(sample)
                )
            )

        X = np.append(X, sample_X, axis=0)
        print(X[np.argmin(y)])
        y = np.array([f(x) for x in X]).flatten()

    if optimium is not None:
        optimium = np.asarray(optimium)
        representative_sample = X[np.argmin(y)]
        np.testing.assert_array_less(np.abs(representative_sample - optimium), atol)

    return X, y


def test_squiggle_explores_parameter_space():
    # This test checks whether the bayes algorithm correctly explores the parameter space
    # we sample a ton of positive examples, ignoring the negative side
    X = np.random.uniform(0, 5, 200)[:, None]
    Y = squiggle(X.ravel())
    (sample, prob, pred, _, _,) = bayes.next_sample(
        sample_X=X, sample_y=Y, X_bounds=[[-5.0, 5.0]], improvement=1.0
    )
    assert sample[0] < 0.0, "Greater than 0 {}".format(sample[0])
    # we sample missing a big chunk between 1 and 3
    X = np.vstack(
        (np.random.uniform(0, 1, 200)[:, None], np.random.uniform(4, 5, 200)[:, None])
    )
    Y = squiggle(X.ravel())
    (
        sample,
        prob,
        pred,
        _,
        _,
    ) = bayes.next_sample(sample_X=X, sample_y=Y, X_bounds=[[0.0, 5.0]])
    assert (
        sample[0] > 1.0 and sample[0] < 4.0
    ), "Sample outside of 1-3 range: {}".format(sample[0])


def test_squiggle_convergence():
    x_init = np.random.uniform(0, 5, 1)[:, None]
    run_iterations(squiggle, [[0.0, 5.0]], 200, x_init, optimium=[3.6], atol=0.2)


def test_squiggle_convergence_to_maximum():
    # This test checks whether the bayes algorithm correctly explores the parameter space
    # we sample a ton of positive examples, ignoring the negative side
    def f(x):
        return -squiggle(x)

    x_init = np.random.uniform(0, 5, 1)[:, None]
    run_iterations(f, [[0.0, 5.0]], 200, x_init, optimium=[2], atol=0.2)


def test_nans():
    def f(x):
        return np.zeros_like(x) * np.nan

    X = np.random.uniform(0, 5, 200)[:, None]
    new_x, new_y = run_iterations(f, [[-10, 10]], 1, X)
    assert new_x[-1][0] < 10.0
    assert np.isnan(new_y[-1])
    new_x += np.random.uniform(0, 5, len(new_x))[:, None]
    new_x, new_y = run_iterations(f, [[-10, 10]], 1, X)
    assert new_x[-1][0] < 10.0
    assert np.isnan(new_y[-1])


def test_squiggle_int():
    f = squiggle
    X = np.random.uniform(0, 5, 200)[:, None]
    new_X, new_y = run_iterations(f, [[-10, 10]], 1, X)
    sample = new_X[-1][0]
    assert sample < 0.0, "Greater than 0 {}".format(sample)
    assert np.isclose(sample % 1, 0)


def test_iterations_rosenbrock():
    dimensions = 3
    # x_init = np.random.uniform(0, 2, size=(1, dimensions))
    x_init = np.zeros((1, dimensions))
    run_iterations(
        rosenbrock,
        [[0.0, 2.0]] * dimensions,
        300,
        x_init,
        optimium=[1, 1, 1],
        atol=0.2,
        improvement=0.1,
    )


def test_iterations_squiggle_chunked():
    run_iterations(
        squiggle,
        [[0.0, 5.0]],
        chunk_size=5,
        num_iterations=200,
        optimium=[3.6],
        improvement=0.1,
    )


def test_bayes_search_with_zero_runs_begins_correctly(
    sweep_config_bayes_search_2params_with_metric,
):
    run = next_run(sweep_config_bayes_search_2params_with_metric, [])
    assert isinstance(run.config["v1"]["value"], int) and isinstance(
        run.config["v2"]["value"], float
    )
    v1 = sweep_config_bayes_search_2params_with_metric["parameters"]["v1"]
    v1_min, v1_max = v1["min"], v1["max"]
    v2 = sweep_config_bayes_search_2params_with_metric["parameters"]["v2"]
    v2_min, v2_max = v2["min"], v2["max"]
    assert v1_min <= run.config["v1"]["value"] <= v1_max
    assert v2_min <= run.config["v2"]["value"] <= v2_max


# search with 2 finished runs - hardcoded results
def test_runs_bayes_runs2(sweep_config_bayes_search_2params_with_metric):
    def loss_func(run: SweepRun) -> floating:
        return (run.config["v1"]["value"] - 5) ** 2 + (
            run.config["v2"]["value"] - 2
        ) ** 2

    r1 = SweepRun(
        name="b",
        state=RunState.finished,
        history=[
            {"loss": 5.0},
        ],
        config={"v1": {"value": 7}, "v2": {"value": 6}},
        summary_metrics={"zloss": 1.2},
    )
    r2 = SweepRun(
        name="b2",
        state=RunState.finished,
        config={"v1": {"value": 1}, "v2": {"value": 8}},
        summary_metrics={"loss": 52.0},
        history=[],
    )
    # need two (non running) runs before we get a new set of parameters

    runs = [r1, r2]

    for _ in range(200):
        suggestion = next_run(sweep_config_bayes_search_2params_with_metric, runs)
        metric = {"loss": loss_func(suggestion)}
        suggestion.history = [metric]
        suggestion.state = RunState.finished
        runs.append(suggestion)

    best_run = min(runs, key=lambda r: r.metric_extremum("loss", "minimum"))
    best_x = np.asarray(
        [best_run.config["v1"]["value"], best_run.config["v2"]["value"]]
    )
    optimum = np.asarray([5.0, 2.0])
    np.testing.assert_array_less(np.abs(best_x - optimum), 0.2)


# search with 2 finished runs - hardcoded results - missing metric
def test_runs_bayes_runs2_missingmetric():

    config = SweepConfig(
        {
            "metric": {"name": "loss", "goal": "minimize"},
            "method": "bayes",
            "parameters": {
                "v2": {"min": 1, "max": 10},
            },
        }
    )

    r1 = SweepRun(
        name="b",
        state=RunState.finished,
        history=[
            {"xloss": 5.0},
        ],
        config={"v2": {"value": 6}},
        summary_metrics={"zloss": 1.2},
    )
    r2 = SweepRun(
        name="b2",
        state=RunState.finished,
        config={"v2": {"value": 8}},
        summary_metrics={"xloss": 52.0},
        history=[],
    )

    runs = [r1, r2]
    for _ in range(200):
        suggestion = next_run(config, runs)
        suggestion.state = RunState.finished
        runs.append(suggestion)

    # should just choose random runs in this case as they are all imputed with the same value (zero)
    # for the loss function
    check_that_samples_are_from_the_same_distribution(
        [run.config["v2"]["value"] for run in runs],
        np.random.uniform(1, 10, 202),
        np.linspace(1, 10, 11),
    )


def test_runs_bayes_runs2_missingmetric_acc():

    config = SweepConfig(
        {
            "method": "bayes",
            "metric": {
                "name": "acc",
                "goal": "maximize",
            },
            "parameters": {"v1": {"min": 1, "max": 10}, "v2": {"min": 1, "max": 10}},
        }
    )

    r1 = SweepRun(
        name="b",
        state=RunState.finished,
        history=[
            {"xloss": 5.0},
        ],
        config={"v1": {"value": 7}, "v2": {"value": 6}},
        summary_metrics={"zloss": 1.2},
    )
    r2 = SweepRun(
        name="b2",
        state=RunState.finished,
        config={"v1": {"value": 1}, "v2": {"value": 8}},
        summary_metrics={"xloss": 52.0},
        history=[],
    )

    runs = [r1, r2]
    for _ in range(200):
        suggestion = next_run(config, runs)
        suggestion.state = RunState.finished
        runs.append(suggestion)

    # should just choose random runs in this case as they are all imputed with the same value (zero)
    # for the loss function
    check_that_samples_are_from_the_same_distribution(
        [run.config["v2"]["value"] for run in runs],
        np.random.uniform(1, 10, 202),
        np.linspace(1, 10, 11),
    )


def test_runs_bayes_nan(sweep_config_bayes_search_2params_with_metric):
    r1 = SweepRun(
        name="b",
        state=RunState.finished,
        config={"v1": {"value": 7}, "v2": {"value": 6}},
        summary_metrics={},
        history=[
            {"loss": float("NaN")},
        ],
    )
    r2 = SweepRun(
        name="b",
        state=RunState.finished,
        config={"v1": {"value": 1}, "v2": {"value": 8}},
        summary_metrics={"loss": float("NaN")},
        history=[],
    )
    r3 = SweepRun(
        name="b",
        state=RunState.finished,
        config={"v1": {"value": 2}, "v2": {"value": 3}},
        summary_metrics={},
        history=[
            {"loss": "NaN"},
        ],
    )
    r4 = SweepRun(
        name="b",
        state=RunState.finished,
        config={"v1": {"value": 4}, "v2": {"value": 5}},
        summary_metrics={"loss": "NaN"},
        history=[],
    )
    # need two (non running) runs before we get a new set of parameters
    runs = [r1, r2, r3, r4]

    for _ in range(200):
        suggestion = next_run(sweep_config_bayes_search_2params_with_metric, runs)
        suggestion.state = RunState.finished
        runs.append(suggestion)

    # should just choose random runs in this case as they are all imputed with the same value (zero)
    # for the loss function
    check_that_samples_are_from_the_same_distribution(
        [run.config["v2"]["value"] for run in runs],
        np.random.uniform(1, 10, 202),
        np.linspace(1, 10, 11),
    )


def test_runs_bayes_categorical_list():

    v2_min = 1
    v2_max = 10

    config = {
        "method": "bayes",
        "metric": {
            "name": "acc",
            "goal": "maximize",
        },
        "parameters": {
            "v1": {
                "distribution": "categorical",
                "values": [(2, 3), [3, 4], ["5", "6"], [(7, 8), ["9", [10, 11]]]],
            },
            "v2": {"min": v2_min, "max": v2_max},
        },
    }

    def loss_func(x: SweepRun) -> floating:
        v2_acc = 0.5 * (x.config["v2"]["value"] - v2_min) / (v2_max - v2_min)
        v1_acc = [0.1, 0.2, 0.5, 0.1][
            config["parameters"]["v1"]["values"].index(x.config["v1"]["value"])
        ]
        return v1_acc + v2_acc

    r1 = SweepRun(
        name="b",
        state=RunState.finished,
        config={"v1": {"value": [3, 4]}, "v2": {"value": 5}},
        history=[],
    )
    r1.summary_metrics = {"acc": loss_func(r1)}

    r2 = SweepRun(
        name="b",
        state=RunState.finished,
        config={"v1": {"value": (2, 3)}, "v2": {"value": 5}},
        history=[],
    )
    r2.summary_metrics = {"acc": loss_func(r2)}

    runs = [r1, r2]
    for _ in range(200):
        suggestion = next_run(config, runs)
        metric = {"acc": loss_func(suggestion)}
        suggestion.history = [metric]
        suggestion.state = RunState.finished
        runs.append(suggestion)

    best_run = max(runs, key=lambda r: r.metric_extremum("acc", "maximum"))
    best_x = [best_run.config["v1"]["value"], best_run.config["v2"]["value"]]

    assert best_x[0] == ["5", "6"]
    assert np.abs(best_x[1] - 10) < 0.2


def test_bayes_can_handle_preemptible_or_preempting_runs():

    v2_min = 1
    v2_max = 10

    config = {
        "method": "bayes",
        "metric": {
            "name": "acc",
            "goal": "maximize",
        },
        "parameters": {
            "v1": {
                "distribution": "categorical",
                "values": [(2, 3), [3, 4], ["5", "6"], [(7, 8), ["9", [10, 11]]]],
            },
            "v2": {"min": v2_min, "max": v2_max},
        },
    }

    def loss_func(x: SweepRun) -> floating:
        v2_acc = 0.5 * (x.config["v2"]["value"] - v2_min) / (v2_max - v2_min)
        v1_acc = [0.1, 0.2, 0.5, 0.1][
            config["parameters"]["v1"]["values"].index(x.config["v1"]["value"])
        ]
        return v1_acc + v2_acc

    r1 = SweepRun(
        name="b",
        state=RunState.preempted,
        config={"v1": {"value": [3, 4]}, "v2": {"value": 5}},
        history=[],
    )
    r1.summary_metrics = {"acc": loss_func(r1)}

    # this should not raise, and should produce the same result as if r1.state was running
    seed = np.random.get_state()
    pred = next_run(config, [r1])
    r1.state = RunState.running
    np.random.set_state(seed)
    true = next_run(config, [r1])
    assert pred.config == true.config

    r2 = SweepRun(
        name="b",
        state=RunState.preempting,
        config={"v1": {"value": (2, 3)}, "v2": {"value": 5}},
        history=[],
    )
    r2.summary_metrics = {"acc": loss_func(r2)}

    seed = np.random.get_state()
    pred = next_run(config, [r2])
    r2.state = RunState.running
    np.random.set_state(seed)
    true = next_run(config, [r2])
    assert pred.config == true.config


def test_that_constant_parameters_are_sampled_correctly():
    # from https://console.cloud.google.com/logs/query;cursorTimestamp=2021-07-07T01:02:35.636458264Z;pinnedLogId=2021-07-07T01:02:35.636458264Z%2Fxl7xiuf6empym;query=jsonPayload.message%3D%22Mismatch%20detected%20in%20shadow%20sweep%20provider%22%0AjsonPayload.data.config%20!~%20%22%5C%22method%5C%22:%20%5C%22grid%5C%22%22%0Atimestamp%3D%222021-07-07T01:02:35.636458264Z%22%0AinsertId%3D%22xl7xiuf6empym%22%0Atimestamp%3D%222021-07-07T01:02:35.636458264Z%22%0AinsertId%3D%22xl7xiuf6empym%22?project=wandb-production
    config = {
        "name": "Elastic Sweep 07-06 1027",
        "method": "bayes",
        "metric": {"goal": "maximize", "name": "IC_plus_Hitrate_risk_adj"},
        "parameters": {
            "tol": {"values": [1e-05, 0.0001, 0.0002, 2.5e-05, 5e-05, 7.5e-05]},
            "alpha": {
                "max": 6.907755278982137,
                "min": -9.210340371976182,
                "distribution": "log_uniform",
            },
            "model": {"value": "ElasticNet"},
            "target": {"values": ["1D_spec", "5D_spec", "10D_spec"]},
            "l1_ratio": {"max": 1, "min": 0, "distribution": "uniform"},
            "selection": {"values": ["cyclic", "random"]},
            "decay_days": {"values": [1, 2, 3, 4, 5, 6]},
            "cat_columns": {"value": ["Sector", "Industry", "market_simple_regime"]},
            "num_columns": {
                "value": [
                    "sales_a1.median",
                    "sales_a1.std_dev",
                    "sales_a1.median_slope",
                    "ebit_a1.median",
                    "ebit_a1.std_dev",
                    "ebit_a1.median_slope",
                    "Log Market Cap",
                    "option_skew",
                    "option_3m_90_vol",
                    "option_3m_105_vol",
                ]
            },
            "ppf_target_by": {"values": [None, "Sector", "Industry"]},
            "ppf_the_target": {"value": True},
        },
    }

    null = None
    true = True
    false = False
    runs = [
        {
            "name": "cv4uof8s",
            "config": {
                "tol": {"desc": null, "value": 7.5e-05},
                "alpha": {"desc": null, "value": 114.94621156856498},
                "model": {"desc": null, "value": "ElasticNet"},
                "_wandb": {
                    "desc": null,
                    "value": {
                        "t": {
                            "1": [3, 5],
                            "2": [3, 5],
                            "3": [2],
                            "4": "3.7.9",
                            "5": "0.10.31",
                            "8": [1, 5],
                        },
                        "framework": "tensorflow",
                        "cli_version": "0.10.31",
                        "is_jupyter_run": true,
                        "python_version": "3.7.9",
                        "is_kaggle_kernel": false,
                    },
                },
                "target": {"desc": null, "value": "1D_spec"},
                "l1_ratio": {"desc": null, "value": 0.7475487707624942},
                "selection": {"desc": null, "value": "random"},
                "decay_days": {"desc": null, "value": 4},
                "cat_columns": {
                    "desc": null,
                    "value": ["Sector", "Industry", "market_simple_regime"],
                },
                "num_columns": {
                    "desc": null,
                    "value": [
                        "sales_a1.median",
                        "sales_a1.std_dev",
                        "sales_a1.median_slope",
                        "ebit_a1.median",
                        "ebit_a1.std_dev",
                        "ebit_a1.median_slope",
                        "Log Market Cap",
                        "option_skew",
                        "option_3m_90_vol",
                        "option_3m_105_vol",
                    ],
                },
                "ppf_target_by": {"desc": null, "value": "Industry"},
                "ppf_the_target": {"desc": null, "value": true},
            },
            "summaryMetrics": {},
            "state": "failed",
        },
        {
            "config": {
                "tol": {"desc": null, "value": 7.5e-05},
                "alpha": {"desc": null, "value": 0.07239397556453357},
                "model": {"desc": null, "value": "ElasticNet"},
                "_wandb": {
                    "desc": null,
                    "value": {
                        "t": {
                            "1": [3, 5],
                            "2": [3, 5],
                            "3": [2],
                            "4": "3.7.9",
                            "5": "0.10.31",
                            "8": [1, 5],
                        },
                        "framework": "tensorflow",
                        "cli_version": "0.10.31",
                        "is_jupyter_run": true,
                        "python_version": "3.7.9",
                        "is_kaggle_kernel": false,
                    },
                },
                "target": {"desc": null, "value": "1D_spec"},
                "l1_ratio": {"desc": null, "value": 0.9317758724316296},
                "selection": {"desc": null, "value": "random"},
                "decay_days": {"desc": null, "value": 6},
                "cat_columns": {
                    "desc": null,
                    "value": ["Sector", "Industry", "market_simple_regime"],
                },
                "num_columns": {
                    "desc": null,
                    "value": [
                        "sales_a1.median",
                        "sales_a1.std_dev",
                        "sales_a1.median_slope",
                        "ebit_a1.median",
                        "ebit_a1.std_dev",
                        "ebit_a1.median_slope",
                        "Log Market Cap",
                        "option_skew",
                        "option_3m_90_vol",
                        "option_3m_105_vol",
                    ],
                },
                "ppf_target_by": {"desc": null, "value": null},
                "ppf_the_target": {"desc": null, "value": true},
            },
            "summaryMetrics": {},
            "state": "failed",
            "name": "633fjdhd",
        },
        {
            "name": "77ywpu3j",
            "summaryMetrics": {},
            "state": "failed",
            "config": {
                "tol": {"desc": null, "value": 5e-05},
                "alpha": {"desc": null, "value": 0.002774852781906077},
                "model": {"desc": null, "value": "ElasticNet"},
                "_wandb": {
                    "desc": null,
                    "value": {
                        "t": {
                            "1": [3, 5],
                            "2": [3, 5],
                            "3": [2],
                            "4": "3.7.9",
                            "5": "0.10.31",
                            "8": [1, 5],
                        },
                        "framework": "tensorflow",
                        "cli_version": "0.10.31",
                        "is_jupyter_run": true,
                        "python_version": "3.7.9",
                        "is_kaggle_kernel": false,
                    },
                },
                "target": {"desc": null, "value": "10D_spec"},
                "l1_ratio": {"desc": null, "value": 0.01483000099874121},
                "selection": {"desc": null, "value": "cyclic"},
                "decay_days": {"desc": null, "value": 6},
                "cat_columns": {
                    "desc": null,
                    "value": ["Sector", "Industry", "market_simple_regime"],
                },
                "num_columns": {
                    "desc": null,
                    "value": [
                        "sales_a1.median",
                        "sales_a1.std_dev",
                        "sales_a1.median_slope",
                        "ebit_a1.median",
                        "ebit_a1.std_dev",
                        "ebit_a1.median_slope",
                        "Log Market Cap",
                        "option_skew",
                        "option_3m_90_vol",
                        "option_3m_105_vol",
                    ],
                },
                "ppf_target_by": {"desc": null, "value": null},
                "ppf_the_target": {"desc": null, "value": true},
            },
        },
        {
            "name": "06z3nutf",
            "summaryMetrics": {},
            "config": {
                "tol": {"desc": null, "value": 0.0001},
                "alpha": {"desc": null, "value": 0.005228908948292734},
                "model": {"desc": null, "value": "ElasticNet"},
                "_wandb": {
                    "desc": null,
                    "value": {
                        "t": {
                            "1": [3, 5],
                            "2": [3, 5],
                            "3": [2],
                            "4": "3.7.9",
                            "5": "0.10.31",
                            "8": [1, 5],
                        },
                        "framework": "tensorflow",
                        "cli_version": "0.10.31",
                        "is_jupyter_run": true,
                        "python_version": "3.7.9",
                        "is_kaggle_kernel": false,
                    },
                },
                "target": {"desc": null, "value": "1D_spec"},
                "l1_ratio": {"desc": null, "value": 0.974230005151591},
                "selection": {"desc": null, "value": "random"},
                "decay_days": {"desc": null, "value": 1},
                "cat_columns": {
                    "desc": null,
                    "value": ["Sector", "Industry", "market_simple_regime"],
                },
                "num_columns": {
                    "desc": null,
                    "value": [
                        "sales_a1.median",
                        "sales_a1.std_dev",
                        "sales_a1.median_slope",
                        "ebit_a1.median",
                        "ebit_a1.std_dev",
                        "ebit_a1.median_slope",
                        "Log Market Cap",
                        "option_skew",
                        "option_3m_90_vol",
                        "option_3m_105_vol",
                    ],
                },
                "ppf_target_by": {"desc": null, "value": null},
                "ppf_the_target": {"desc": null, "value": true},
            },
            "state": "failed",
        },
        {
            "summaryMetrics": {},
            "config": {
                "tol": {"desc": null, "value": 0.0001},
                "alpha": {"desc": null, "value": 2.2736520853608524},
                "model": {"desc": null, "value": "ElasticNet"},
                "_wandb": {
                    "desc": null,
                    "value": {
                        "t": {
                            "1": [3, 5],
                            "2": [3, 5],
                            "3": [2],
                            "4": "3.7.9",
                            "5": "0.10.31",
                            "8": [1, 5],
                        },
                        "framework": "tensorflow",
                        "cli_version": "0.10.31",
                        "is_jupyter_run": true,
                        "python_version": "3.7.9",
                        "is_kaggle_kernel": false,
                    },
                },
                "target": {"desc": null, "value": "1D_spec"},
                "l1_ratio": {"desc": null, "value": 0.6713230851920869},
                "selection": {"desc": null, "value": "random"},
                "decay_days": {"desc": null, "value": 1},
                "cat_columns": {
                    "desc": null,
                    "value": ["Sector", "Industry", "market_simple_regime"],
                },
                "num_columns": {
                    "desc": null,
                    "value": [
                        "sales_a1.median",
                        "sales_a1.std_dev",
                        "sales_a1.median_slope",
                        "ebit_a1.median",
                        "ebit_a1.std_dev",
                        "ebit_a1.median_slope",
                        "Log Market Cap",
                        "option_skew",
                        "option_3m_90_vol",
                        "option_3m_105_vol",
                    ],
                },
                "ppf_target_by": {"desc": null, "value": "Industry"},
                "ppf_the_target": {"desc": null, "value": true},
            },
            "state": "failed",
            "name": "1ddlc02y",
        },
    ]

    suggestion = next_run(config, [SweepRun(**run) for run in runs])
    for key in suggestion.config:
        if key in config["parameters"] and key != "ppf_target_by":
            assert suggestion.config[key]["value"] is not None


def test_metric_extremum_in_bayes_search():
    # from https://console.cloud.google.com/logs/query;query=ygnwe8ptupj33get%0A;timeRange=2021-08-03T21:34:50.082Z%2F2021-08-03T21:34:59.082Z;summaryFields=:false:32:beginning;cursorTimestamp=2021-08-03T21:34:51.189649752Z?project=wandb-production
    data_path = f"{os.path.dirname(__file__)}/data/ygnwe8ptupj33get.decoded.json"
    with open(data_path, "r") as f:
        data = json.load(f)
    _, _, _, y = bayes._construct_gp_data(
        [SweepRun(**r) for r in data["jsonPayload"]["data"]["runs"]],
        data["jsonPayload"]["data"]["config"],
    )
    np.testing.assert_array_less(np.abs(y + 98), 5)



#
from __future__ import absolute_import

import json
import platform
import subprocess
import threading
import time
from typing import Dict, List, Optional, Union

import psutil
import wandb
from wandb import util
from wandb.vendor.pynvml import pynvml

from . import tpu
from ..interface.interface import BackendSender
from ..lib import telemetry


GPUHandle = object
SamplerDict = Dict[str, List[float]]
StatsDict = Dict[str, Union[float, Dict[str, float]]]


# TODO: hard coded max watts as 16.5, found this number in the SMC list.
# Eventually we can have the apple_gpu_stats binary query for this.
M1_MAX_POWER_WATTS = 16.5


def gpu_in_use_by_this_process(gpu_handle: GPUHandle) -> bool:
    if not psutil:
        return False

    # NOTE: this optimizes for the case where wandb was initialized from
    # iniside the user script (i.e. `wandb.init()`). If we ran using
    # `wandb run` on the command line, the shell will be detected as the
    # parent, possible resulting in sibling processes being incorrectly
    # indentified as part of this process -- still better than not
    # detecting in-use gpus at all.
    base_process = psutil.Process().parent() or psutil.Process()

    our_processes = base_process.children(recursive=True)
    our_processes.append(base_process)

    our_pids = set([process.pid for process in our_processes])

    compute_pids = set(
        [
            process.pid
            for process in pynvml.nvmlDeviceGetComputeRunningProcesses(gpu_handle)
        ]
    )
    graphics_pids = set(
        [
            process.pid
            for process in pynvml.nvmlDeviceGetGraphicsRunningProcesses(gpu_handle)
        ]
    )

    pids_using_device = compute_pids | graphics_pids

    return len(pids_using_device & our_pids) > 0


class SystemStats(object):

    _pid: int
    _interface: BackendSender
    sampler: SamplerDict
    samples: int
    _thread: Optional[threading.Thread]
    gpu_count: int

    def __init__(self, pid: int, interface: BackendSender) -> None:
        try:
            pynvml.nvmlInit()
            self.gpu_count = pynvml.nvmlDeviceGetCount()
        except pynvml.NVMLError:
            self.gpu_count = 0
        # self.run = run
        self._pid = pid
        self._interface = interface
        self.sampler = {}
        self.samples = 0
        self._shutdown = False
        self._telem = telemetry.TelemetryRecord()
        if psutil:
            net = psutil.net_io_counters()
            self.network_init = {"sent": net.bytes_sent, "recv": net.bytes_recv}
        else:
            wandb.termlog(
                "psutil not installed, only GPU stats will be reported.  Install with pip install psutil"
            )
        self._thread = None
        self._tpu_profiler = None

        if tpu.is_tpu_available():
            try:
                self._tpu_profiler = tpu.get_profiler()
            except Exception as e:
                wandb.termlog("Error initializing TPUProfiler: " + str(e))

    def start(self) -> None:
        if self._thread is None:
            self._shutdown = False
            self._thread = threading.Thread(target=self._thread_body)
            self._thread.daemon = True
        if not self._thread.is_alive():
            self._thread.start()
        if self._tpu_profiler:
            self._tpu_profiler.start()

    @property
    def proc(self) -> psutil.Process:
        return psutil.Process(pid=self._pid)

    @property
    def sample_rate_seconds(self) -> float:
        """Sample system stats every this many seconds, defaults to 2, min is 0.5"""
        return 2
        # return max(0.5, self._api.dynamic_settings["system_sample_seconds"])

    @property
    def samples_to_average(self) -> int:
        """The number of samples to average before pushing, defaults to 15 valid range (2:30)"""
        return 15
        # return min(30, max(2, self._api.dynamic_settings["system_samples"]))

    def _thread_body(self) -> None:
        while True:
            stats = self.stats()
            for stat, value in stats.items():
                if isinstance(value, (int, float)):
                    self.sampler[stat] = self.sampler.get(stat, [])
                    self.sampler[stat].append(value)
            self.samples += 1
            if self._shutdown or self.samples >= self.samples_to_average:
                self.flush()
                if self._shutdown:
                    break
            seconds = 0.0
            while seconds < self.sample_rate_seconds:
                time.sleep(0.1)
                seconds += 0.1
                if self._shutdown:
                    self.flush()
                    return

    def shutdown(self) -> None:
        self._shutdown = True
        try:
            if self._thread is not None:
                self._thread.join()
        finally:
            self._thread = None
        if self._tpu_profiler:
            self._tpu_profiler.stop()

    def flush(self) -> None:
        stats = self.stats()
        for stat, value in stats.items():
            # TODO: a bit hacky, we assume all numbers should be averaged.  If you want
            # max for a stat, you must put it in a sub key, like ["network"]["sent"]
            if isinstance(value, (float, int)):
                # samples = list(self.sampler.get(stat, [stats[stat]]))
                samples = list(self.sampler.get(stat, [value]))
                stats[stat] = round(sum(samples) / len(samples), 2)
        # self.run.events.track("system", stats, _wandb=True)
        if self._interface:
            self._interface.publish_stats(stats)
        self.samples = 0
        self.sampler = {}

    def stats(self) -> StatsDict:
        stats: StatsDict = {}
        for i in range(0, self.gpu_count):
            handle = pynvml.nvmlDeviceGetHandleByIndex(i)
            try:
                utilz = pynvml.nvmlDeviceGetUtilizationRates(handle)
                memory = pynvml.nvmlDeviceGetMemoryInfo(handle)
                temp = pynvml.nvmlDeviceGetTemperature(
                    handle, pynvml.NVML_TEMPERATURE_GPU
                )
                in_use_by_us = gpu_in_use_by_this_process(handle)

                stats["gpu.{}.{}".format(i, "gpu")] = utilz.gpu
                stats["gpu.{}.{}".format(i, "memory")] = utilz.memory
                stats["gpu.{}.{}".format(i, "memoryAllocated")] = (
                    memory.used / float(memory.total)
                ) * 100
                stats["gpu.{}.{}".format(i, "temp")] = temp

                if in_use_by_us:
                    stats["gpu.process.{}.{}".format(i, "gpu")] = utilz.gpu
                    stats["gpu.process.{}.{}".format(i, "memory")] = utilz.memory
                    stats["gpu.process.{}.{}".format(i, "memoryAllocated")] = (
                        memory.used / float(memory.total)
                    ) * 100
                    stats["gpu.process.{}.{}".format(i, "temp")] = temp

                    # Some GPUs don't provide information about power usage
                try:
                    power_watts = pynvml.nvmlDeviceGetPowerUsage(handle) / 1000.0
                    power_capacity_watts = (
                        pynvml.nvmlDeviceGetEnforcedPowerLimit(handle) / 1000.0
                    )
                    power_usage = (power_watts / power_capacity_watts) * 100

                    stats["gpu.{}.{}".format(i, "powerWatts")] = power_watts
                    stats["gpu.{}.{}".format(i, "powerPercent")] = power_usage

                    if in_use_by_us:
                        stats["gpu.process.{}.{}".format(i, "powerWatts")] = power_watts
                        stats[
                            "gpu.process.{}.{}".format(i, "powerPercent")
                        ] = power_usage

                except pynvml.NVMLError:
                    pass

            except pynvml.NVMLError:
                pass

        # On Apple M1 systems let's look for the gpu
        if (
            platform.system() == "Darwin"
            and platform.processor() == "arm"
            and self.gpu_count == 0
        ):
            try:
                out = subprocess.check_output([util.apple_gpu_stats_binary(), "--json"])
                m1_stats = json.loads(out.split(b"\n")[0])
                stats["gpu.0.gpu"] = m1_stats["utilization"]
                stats["gpu.0.memoryAllocated"] = m1_stats["mem_used"]
                stats["gpu.0.temp"] = m1_stats["temperature"]
                stats["gpu.0.powerWatts"] = m1_stats["power"]
                stats["gpu.0.powerPercent"] = (
                    m1_stats["power"] / M1_MAX_POWER_WATTS
                ) * 100
                # TODO: this stat could be useful eventually, it was consistently
                # 0 in my experimentation and requires a frontend change
                # so leaving it out for now.
                # stats["gpu.0.cpuWaitMs"] = m1_stats["cpu_wait_ms"]

                if self._interface and not self._telem.env.m1_gpu:
                    self._telem.env.m1_gpu = True
                    self._interface.publish_telemetry(self._telem)

            except (OSError, ValueError, TypeError, subprocess.CalledProcessError) as e:
                wandb.termwarn("GPU stats error {}".format(e))
                pass

        if psutil:
            net = psutil.net_io_counters()
            sysmem = psutil.virtual_memory()
            stats["cpu"] = psutil.cpu_percent()
            stats["memory"] = sysmem.percent
            stats["network"] = {
                "sent": net.bytes_sent - self.network_init["sent"],
                "recv": net.bytes_recv - self.network_init["recv"],
            }
            # TODO: maybe show other partitions, will likely need user to configure
            stats["disk"] = psutil.disk_usage("/").percent
            stats["proc.memory.availableMB"] = sysmem.available / 1048576.0
            try:
                stats["proc.memory.rssMB"] = self.proc.memory_info().rss / 1048576.0
                stats["proc.memory.percent"] = self.proc.memory_percent()
                stats["proc.cpu.threads"] = self.proc.num_threads()
            except psutil.NoSuchProcess:
                pass
        if self._tpu_profiler:
            stats["tpu"] = self._tpu_profiler.get_tpu_utilization()
        return stats



import os
import sys
import time
from typing import Any, Dict, Iterable, List, Optional

import wandb
from wandb.apis.internal import Api
from wandb.errors import LaunchError
import wandb.util as util

from .._project_spec import create_project_from_spec, fetch_and_validate_project
from ..runner.abstract import AbstractRun, State
from ..runner.loader import load_backend
from ..utils import (
    _is_wandb_local_uri,
    PROJECT_DOCKER_ARGS,
    PROJECT_SYNCHRONOUS,
)

AGENT_POLLING_INTERVAL = 10


def _convert_access(access: str) -> str:
    """Converts access string to a value accepted by wandb."""
    access = access.upper()
    assert (
        access == "PROJECT" or access == "USER"
    ), "Queue access must be either project or user"
    return access


class LaunchAgent(object):
    """Launch agent class which polls run given run queues and launches runs for wandb launch."""

    STATE_MAP: Dict[str, State] = {}

    def __init__(self, entity: str, project: str, queues: Iterable[str] = None):
        self._entity = entity
        self._project = project
        self._api = Api()
        self._settings = wandb.Settings()
        self._base_url = self._api.settings().get("base_url")
        self._jobs: Dict[int, AbstractRun] = {}
        self._ticks = 0
        self._running = 0
        self._cwd = os.getcwd()
        self._namespace = wandb.util.generate_id()
        self._access = _convert_access("project")
        self._queues: List[str] = []
        self.setup_run_queues(queues)

    def setup_run_queues(self, queues: Optional[Iterable[str]]) -> None:
        """Checks the project to ensure run queues exist then adds them to a list to be watched by the agent."""
        # TODO: add run queue filtering to server
        project_run_queues = self._api.get_project_run_queues(
            self._entity, self._project
        )
        existing_run_queue_names = set(
            [run_queue["name"] for run_queue in project_run_queues]
        )
        if queues is None:
            for queue in project_run_queues:
                if queue["name"] == "default":
                    self._queues = ["default"]
                    return
            raise LaunchError(
                "Error launching launch-agent: Requested default queue for {}/{} but default queue does not exist.".format(
                    self._entity, self._project
                )
            )
        else:
            for queue in queues:
                if queue not in existing_run_queue_names:
                    raise LaunchError(
                        "Error launching launch-agent: {} does not exist for {}/{}".format(
                            queue, self._entity, self._project
                        )
                    )
                self._queues.append(queue)

    @property
    def job_ids(self) -> List[int]:
        """Returns a list of keys running job ids for the agent."""
        return list(self._jobs.keys())

    def pop_from_queue(self, queue: str) -> Any:
        """Pops an item off the runqueue to run as a job."""
        try:
            ups = self._api.pop_from_run_queue(
                queue, entity=self._entity, project=self._project
            )
        except Exception as e:
            print("Exception:", e)
            return None
        return ups

    def print_status(self) -> None:
        """Prints the current status of the agent."""
        print(
            "polling on project {}, queues {} for jobs".format(
                self._project, " ".join(self._queues)
            )
        )

    def finish_job_id(self, job_id: int) -> None:
        """Removes the job from our list for now."""
        # TODO:  keep logs or something for the finished jobs
        del self._jobs[job_id]
        self._running -= 1

    def _update_finished(self, job_id: int) -> None:
        """Check our status enum."""
        if self._jobs[job_id].get_status().state in ["failed", "finished"]:
            self.finish_job_id(job_id)

    def run_job(self, job: Dict[str, Any]) -> None:
        """Sets up project and runs the job."""
        # TODO: logger
        print("agent: got job", job)
        # parse job
        launch_spec = job["runSpec"]
        if launch_spec.get("overrides") and isinstance(
            launch_spec["overrides"].get("args"), list
        ):
            launch_spec["overrides"]["args"] = util._user_args_to_dict(
                launch_spec["overrides"].get("args", [])
            )
        project = create_project_from_spec(launch_spec, self._api)
        project = fetch_and_validate_project(project, self._api)

        resource = launch_spec.get("resource") or "local"
        backend_config: Dict[str, Any] = {
            PROJECT_DOCKER_ARGS: {},
            PROJECT_SYNCHRONOUS: True,
        }
        if _is_wandb_local_uri(self._base_url):
            if sys.platform == "win32":
                backend_config[PROJECT_DOCKER_ARGS]["net"] = "host"
            else:
                backend_config[PROJECT_DOCKER_ARGS]["network"] = "host"
            if sys.platform == "linux" or sys.platform == "linux2":
                backend_config[PROJECT_DOCKER_ARGS][
                    "add-host"
                ] = "host.docker.internal:host-gateway"

        backend_config["runQueueItemId"] = job["runQueueItemId"]
        backend = load_backend(resource, self._api, backend_config)
        backend.verify()

        run = backend.run(project)
        if run:
            self._jobs[run.id] = run
            self._running += 1

    def loop(self) -> None:
        """Main loop function for agent."""
        wandb.termlog(
            "launch agent polling project {}/{} on queues: {}".format(
                self._entity, self._project, ",".join(self._queues)
            )
        )
        try:
            while True:
                self._ticks += 1
                for queue in self._queues:
                    job = self.pop_from_queue(queue)
                    if job:
                        break
                if not job:
                    time.sleep(AGENT_POLLING_INTERVAL)
                    # for job_id in self.job_ids:
                    # self._update_finished(job_id)
                    if self._ticks % 2 == 0:
                        self.print_status()
                    continue
                self.run_job(job)
        except KeyboardInterrupt:
            wandb.termlog("Shutting down, active jobs:")
            self.print_status()



#
# -*- coding: utf-8 -*-
"""
apikey util.
"""

import getpass
import os
import stat
import sys
import textwrap

import requests
import wandb
from wandb.apis import InternalApi
from wandb.errors import term
from wandb.util import _is_databricks, isatty, prompt_choices


LOGIN_CHOICE_ANON = "Private W&B dashboard, no account required"
LOGIN_CHOICE_NEW = "Create a W&B account"
LOGIN_CHOICE_EXISTS = "Use an existing W&B account"
LOGIN_CHOICE_DRYRUN = "Don't visualize my results"
LOGIN_CHOICE_NOTTY = "Unconfigured"
LOGIN_CHOICES = [
    LOGIN_CHOICE_ANON,
    LOGIN_CHOICE_NEW,
    LOGIN_CHOICE_EXISTS,
    LOGIN_CHOICE_DRYRUN,
]


def _fixup_anon_mode(default):
    # Convert weird anonymode values from legacy settings files
    # into one of our expected values.
    anon_mode = default or "never"
    mapping = {"true": "allow", "false": "never"}
    return mapping.get(anon_mode, anon_mode)


def prompt_api_key(  # noqa: C901
    settings,
    api=None,
    input_callback=None,
    browser_callback=None,
    no_offline=False,
    no_create=False,
    local=False,
):
    """Prompt for api key.

    Returns:
        str - if key is configured
        None - if dryrun is selected
        False - if unconfigured (notty)
    """
    input_callback = input_callback or getpass.getpass
    log_string = term.LOG_STRING
    api = api or InternalApi(settings)
    anon_mode = _fixup_anon_mode(settings.anonymous)
    jupyter = settings._jupyter or False
    app_url = api.app_url

    choices = [choice for choice in LOGIN_CHOICES]
    if anon_mode == "never":
        # Omit LOGIN_CHOICE_ANON as a choice if the env var is set to never
        choices.remove(LOGIN_CHOICE_ANON)
    if jupyter or no_offline:
        choices.remove(LOGIN_CHOICE_DRYRUN)
    if jupyter or no_create:
        choices.remove(LOGIN_CHOICE_NEW)

    if jupyter and "google.colab" in sys.modules:
        log_string = term.LOG_STRING_NOCOLOR
        key = wandb.jupyter.attempt_colab_login(app_url)
        if key is not None:
            write_key(settings, key, api=api)
            return key

    if anon_mode == "must":
        result = LOGIN_CHOICE_ANON
    # If we're not in an interactive environment, default to dry-run.
    elif (
        not jupyter and (not isatty(sys.stdout) or not isatty(sys.stdin))
    ) or _is_databricks():
        result = LOGIN_CHOICE_NOTTY
    elif local:
        result = LOGIN_CHOICE_EXISTS
    elif len(choices) == 1:
        result = choices[0]
    else:
        result = prompt_choices(choices)

    api_ask = "%s: Paste an API key from your profile and hit enter: " % log_string
    if result == LOGIN_CHOICE_ANON:
        key = api.create_anonymous_api_key()

        write_key(settings, key, api=api, anonymous=True)
        return key
    elif result == LOGIN_CHOICE_NEW:
        key = browser_callback(signup=True) if browser_callback else None

        if not key:
            wandb.termlog(
                "Create an account here: {}/authorize?signup=true".format(app_url)
            )
            key = input_callback(api_ask).strip()

        write_key(settings, key, api=api)
        return key
    elif result == LOGIN_CHOICE_EXISTS:
        key = browser_callback() if browser_callback else None

        if not key:
            wandb.termlog(
                "You can find your API key in your browser here: {}/authorize".format(
                    app_url
                )
            )
            key = input_callback(api_ask).strip()
        write_key(settings, key, api=api)
        return key
    elif result == LOGIN_CHOICE_NOTTY:
        # TODO: Needs refactor as this needs to be handled by caller
        return False
    else:
        # Jupyter environments don't have a tty, but we can still try logging in using
        # the browser callback if one is supplied.
        key, anonymous = (
            browser_callback() if jupyter and browser_callback else (None, False)
        )

        write_key(settings, key, api=api)
        return key


def write_netrc(host, entity, key):
    """Add our host and key to .netrc"""
    key_prefix, key_suffix = key.split("-", 1) if "-" in key else ("", key)
    if len(key_suffix) != 40:
        wandb.termerror(
            "API-key must be exactly 40 characters long: {} ({} chars)".format(
                key_suffix, len(key_suffix)
            )
        )
        return None
    try:
        normalized_host = host.rstrip("/").split("/")[-1].split(":")[0]
        if normalized_host != "localhost" and "." not in normalized_host:
            wandb.termerror("Host must be a url in the form https://some.address.com")
            return None
        wandb.termlog(
            "Appending key for {} to your netrc file: {}".format(
                normalized_host, os.path.expanduser("~/.netrc")
            )
        )
        machine_line = "machine %s" % normalized_host
        path = os.path.expanduser("~/.netrc")
        orig_lines = None
        try:
            with open(path) as f:
                orig_lines = f.read().strip().split("\n")
        except IOError:
            pass
        with open(path, "w") as f:
            if orig_lines:
                # delete this machine from the file if it's already there.
                skip = 0
                for line in orig_lines:
                    # we fix invalid netrc files with an empty host that we wrote before
                    # verifying host...
                    if line == "machine " or machine_line in line:
                        skip = 2
                    elif skip:
                        skip -= 1
                    else:
                        f.write("%s\n" % line)
            f.write(
                textwrap.dedent(
                    """\
            machine {host}
              login {entity}
              password {key}
            """
                ).format(host=normalized_host, entity=entity, key=key)
            )
        os.chmod(os.path.expanduser("~/.netrc"), stat.S_IRUSR | stat.S_IWUSR)
        return True
    except IOError:
        wandb.termerror("Unable to read ~/.netrc")
        return None


def write_key(settings, key, api=None, anonymous=False):
    if not key:
        return

    # TODO(jhr): api shouldn't be optional or it shouldnt be passed, clean up callers
    api = api or InternalApi()

    # Normal API keys are 40-character hex strings. Onprem API keys have a
    # variable-length prefix, a dash, then the 40-char string.
    prefix, suffix = key.split("-", 1) if "-" in key else ("", key)

    if len(suffix) != 40:
        raise ValueError("API key must be 40 characters long, yours was %s" % len(key))

    if anonymous:
        api.set_setting("anonymous", "true", globally=True, persist=True)
    else:
        api.clear_setting("anonymous", globally=True, persist=True)

    write_netrc(settings.base_url, "user", key)


def api_key(settings=None):
    if not settings:
        settings = wandb.setup().settings
    if settings.api_key:
        return settings.api_key
    auth = requests.utils.get_netrc_auth(settings.base_url)
    if auth:
        return auth[-1]
    return None



from typing import List, Optional, Union, Any, Dict
from enum import Enum
from numbers import Number
import numpy as np
import datetime

from pydantic import BaseModel, Field
from .config import SweepConfig
from ._types import floating


class RunState(str, Enum):
    pending = "pending"
    running = "running"
    finished = "finished"
    killed = "killed"
    crashed = "crashed"
    failed = "failed"
    preempted = "preempted"
    preempting = "preempting"


def is_number(x: Any) -> bool:
    """Check if a value is a finite number."""
    try:
        return np.isscalar(x) and np.isfinite(x) and isinstance(x, Number)
    except TypeError:
        return False


class SweepRun(BaseModel):
    """A wandb Run that is part of a Sweep.

    >>> run = SweepRun(
    ...   name="my_run",
    ...   state=RunState.running,
    ...   config={"a": {"value": 1}},
    ... )

    Args:
        name: Name of the run.
        state: State of the run.
        config: `dict` representation of the run's wandb.config.
        summaryMetrics: `dict` of summary statistics for the run.
        history: List of dicts containing the arguments to calls of wandb.log made during the run.
        search_info: Dict containing information produced by the search algorithm.
        early_terminate_info: Dict containing information produced by the early terminate algorithm.
        stopped: Whether the run was stopped in the sweep
        shouldStop: Whether the run should stop in the sweep
        heartbeat_at: The last time the backend received a heart beat from the run
        exitcode: The exitcode of the process that trained the run
        running: Whether the run is currently running
    """

    name: Optional[str] = None
    summary_metrics: Optional[dict] = Field(
        default_factory=lambda: {}, alias="summaryMetrics"
    )
    history: List[dict] = Field(default_factory=lambda: [], alias="sampledHistory")
    config: dict = Field(default_factory=lambda: {})
    state: RunState = RunState.pending
    search_info: Optional[Dict] = None
    early_terminate_info: Optional[Dict] = None
    stopped: bool = False
    should_stop: bool = Field(default=False, alias="shouldStop")
    heartbeat_at: Optional[datetime.datetime] = Field(default=None, alias="heartbeatAt")
    exitcode: Optional[int] = None
    running: Optional[bool] = None

    class Config:
        use_enum_values = True
        allow_population_by_field_name = True

    def metric_history(
        self, metric_name: str, filter_invalid: bool = False
    ) -> List[floating]:
        return [
            d[metric_name]
            for d in self.history
            if metric_name in d
            and not (filter_invalid and not is_number(d[metric_name]))
        ]

    def summary_metric(self, metric_name: str) -> floating:
        if self.summary_metrics is None:
            raise ValueError("this run has no summary metrics")
        if metric_name not in self.summary_metrics:
            raise KeyError(f"{metric_name} is not a summary metric of this run.")
        return self.summary_metrics[metric_name]

    def metric_extremum(self, metric_name: str, kind: str) -> floating:
        """Calculate the maximum or minimum value of a specified metric.

        >>> run = SweepRun(history=[{'a': 1}, {'b': 3}, {'a': 2, 'b': 4}], summary_metrics={'a': 50})
        >>> assert run.metric_extremum('a', 'maximum') == 50

        Args:
            metric_name: The name of the target metric.
            kind: What kind of extremum to get (either "maximum" or "minimum").

        Returns:
            The maximum or minimum metric.
        """

        cmp_func = np.max if kind == "maximum" else np.min
        try:
            summary_metric = [self.summary_metric(metric_name)]
        except KeyError:
            summary_metric = []
        all_metrics = self.metric_history(metric_name) + summary_metric

        if len(all_metrics) == 0:
            raise ValueError(f"Cannot extract metric {metric_name} from run")

        all_metrics = list(filter(is_number, all_metrics))

        if len(all_metrics) == 0:
            raise ValueError("Run does not have any finite metric values")

        return cmp_func(all_metrics)


def next_run(
    sweep_config: Union[dict, SweepConfig],
    runs: List[SweepRun],
    validate: bool = False,
    **kwargs,
) -> Optional[SweepRun]:
    """Calculate the next run in a sweep.

    >>> suggested_run = next_run({
    ...    'method': 'grid',
    ...    'parameters': {'a': {'values': [1, 2, 3]}}
    ... }, [])
    >>> assert suggested_run.config['a']['value'] == 1

    Args:
        sweep_config: The config for the sweep.
        runs: List of runs in the sweep.
        validate: Whether to validate `sweep_config` against the SweepConfig JSONschema.
           If true, will raise a Validation error if `sweep_config` does not conform to
           the schema. If false, will attempt to run the sweep with an unvalidated schema.

    Returns:
        The suggested run.
    """

    from .grid_search import grid_search_next_run
    from .random_search import random_search_next_run
    from .bayes_search import bayes_search_next_run

    # validate the sweep config
    if validate:
        sweep_config = SweepConfig(sweep_config)

    if "method" not in sweep_config:
        raise ValueError("Sweep config must contain method section")

    if "parameters" not in sweep_config:
        raise ValueError("Sweep config must contain parameters section")

    if not (
        isinstance(sweep_config["parameters"], dict)
        and len(sweep_config["parameters"]) > 0
    ):
        raise ValueError(
            "Parameters section of sweep config must be a dict of at least length 1"
        )

    method = sweep_config["method"]

    if method == "grid":
        return grid_search_next_run(runs, sweep_config, validate=validate, **kwargs)
    elif method == "random":
        return random_search_next_run(sweep_config, validate=validate)
    elif method == "bayes":
        return bayes_search_next_run(runs, sweep_config, validate=validate, **kwargs)
    else:
        raise ValueError(
            f'Invalid search type {method}, must be one of ["grid", "random", "bayes"]'
        )


def stop_runs(
    sweep_config: Union[dict, SweepConfig],
    runs: List[SweepRun],
    validate: bool = False,
) -> List[SweepRun]:
    """Calculate the runs in a sweep to stop by early termination.

    >>> to_stop = stop_runs({
    ...    "method": "grid",
    ...    "metric": {"name": "loss", "goal": "minimize"},
    ...    "early_terminate": {
    ...        "type": "hyperband",
    ...        "max_iter": 5,
    ...        "eta": 2,
    ...        "s": 2,
    ...    },
    ...    "parameters": {"a": {"values": [1, 2, 3]}},
    ... }, [
    ...    SweepRun(
    ...        name="a",
    ...        state=RunState.finished,  # This is already stopped
    ...        history=[
    ...            {"loss": 10},
    ...            {"loss": 9},
    ...        ],
    ...    ),
    ...    SweepRun(
    ...        name="b",
    ...        state=RunState.running,  # This should be stopped
    ...        history=[
    ...            {"loss": 10},
    ...            {"loss": 10},
    ...        ],
    ...    ),
    ...    SweepRun(
    ...        name="c",
    ...        state=RunState.running,  # This passes band 1 but not band 2
    ...        history=[
    ...            {"loss": 10},
    ...            {"loss": 8},
    ...            {"loss": 8},
    ...        ],
    ...    ),
    ...    SweepRun(
    ...        name="d",
    ...        state=RunState.running,
    ...        history=[
    ...            {"loss": 10},
    ...            {"loss": 7},
    ...            {"loss": 7},
    ...        ],
    ...    ),
    ...    SweepRun(
    ...        name="e",
    ...        state=RunState.finished,
    ...        history=[
    ...            {"loss": 10},
    ...            {"loss": 6},
    ...            {"loss": 6},
    ...        ],
    ...    ),
    ...])


    Args:
        sweep_config: The config for the sweep.
        runs: List of runs in the sweep.
        validate: Whether to validate `sweep_config` against the SweepConfig JSONschema.
           If true, will raise a Validation error if `sweep_config` does not conform to
           the schema. If false, will attempt to run the sweep with an unvalidated schema.


    Returns:
        A list of the runs to stop.
    """

    from .hyperband_stopping import hyperband_stop_runs

    # validate the sweep config
    if validate:
        sweep_config = SweepConfig(sweep_config)

    if "metric" not in sweep_config:
        raise ValueError('early terminate requires "metric" section')

    if "early_terminate" not in sweep_config:
        raise ValueError('early terminate requires "early_terminate" section.')
    et_type = sweep_config["early_terminate"]["type"]

    if et_type == "hyperband":
        return hyperband_stop_runs(runs, sweep_config, validate=validate)
    else:
        raise ValueError(
            f'Invalid early stopping type {et_type}, must be one of ["hyperband"]'
        )



# heavily inspired by https://github.com/mlflow/mlflow/blob/master/mlflow/projects/utils.py
import logging
import os
import re
import subprocess
from typing import Any, Dict, Optional, Tuple

import wandb
from wandb import util
from wandb.apis.internal import Api
from wandb.errors import CommError, ExecutionError, LaunchError


# TODO: this should be restricted to just Git repos and not S3 and stuff like that
_GIT_URI_REGEX = re.compile(r"^[^/|^~|^\.].*(git|bitbucket)")
_WANDB_URI_REGEX = re.compile(r"^https://(api.)?wandb")
_WANDB_QA_URI_REGEX = re.compile(
    r"^https?://ap\w.qa.wandb"
)  # for testing, not sure if we wanna keep this
_WANDB_DEV_URI_REGEX = re.compile(
    r"^https?://ap\w.wandb.test"
)  # for testing, not sure if we wanna keep this
_WANDB_LOCAL_DEV_URI_REGEX = re.compile(
    r"^https?://localhost"
)  # for testing, not sure if we wanna keep this


PROJECT_SYNCHRONOUS = "SYNCHRONOUS"
PROJECT_DOCKER_ARGS = "DOCKER_ARGS"

UNCATEGORIZED_PROJECT = "uncategorized"


_logger = logging.getLogger(__name__)


def _is_wandb_uri(uri: str) -> bool:
    return (
        _WANDB_URI_REGEX.match(uri)
        or _WANDB_DEV_URI_REGEX.match(uri)
        or _WANDB_LOCAL_DEV_URI_REGEX.match(uri)
        or _WANDB_QA_URI_REGEX.match(uri)
    ) is not None


def _is_wandb_dev_uri(uri: str) -> bool:
    return bool(_WANDB_DEV_URI_REGEX.match(uri))


def _is_wandb_local_uri(uri: str) -> bool:
    return bool(_WANDB_LOCAL_DEV_URI_REGEX.match(uri))


def _is_git_uri(uri: str) -> bool:
    return bool(_GIT_URI_REGEX.match(uri))


def set_project_entity_defaults(
    uri: str,
    api: Api,
    project: Optional[str],
    entity: Optional[str],
    launch_config: Optional[Dict[str, Any]],
) -> Tuple[str, str]:
    # set the target project and entity if not provided
    if _is_wandb_uri(uri):
        _, uri_project, _ = parse_wandb_uri(uri)
    elif _is_git_uri(uri):
        uri_project = os.path.splitext(os.path.basename(uri))[0]
    else:
        uri_project = UNCATEGORIZED_PROJECT
    if project is None:
        config_project = None
        if launch_config:
            config_project = launch_config.get("project")
        project = config_project or uri_project or UNCATEGORIZED_PROJECT
        wandb.termlog(
            "Target project for this run not specified, defaulting to project {}".format(
                project
            )
        )
    if entity is None:
        config_entity = None
        if launch_config:
            config_entity = launch_config.get("entity")
        entity = config_entity or api.default_entity
        wandb.termlog(
            "Target entity for this run not specified, defaulting to current logged-in user {}".format(
                entity
            )
        )
    return project, entity


def construct_launch_spec(
    uri: str,
    api: Api,
    name: Optional[str],
    project: Optional[str],
    entity: Optional[str],
    docker_image: Optional[str],
    entry_point: Optional[str],
    version: Optional[str],
    parameters: Optional[Dict[str, Any]],
    launch_config: Optional[Dict[str, Any]],
) -> Dict[str, Any]:
    """Constructs the launch specification from CLI arguments."""
    # override base config (if supplied) with supplied args
    launch_spec = launch_config if launch_config is not None else {}
    launch_spec["uri"] = uri
    project, entity = set_project_entity_defaults(
        uri, api, project, entity, launch_config,
    )
    launch_spec["entity"] = entity

    launch_spec["project"] = project
    if name:
        launch_spec["name"] = name
    if "docker" not in launch_spec:
        launch_spec["docker"] = {}
    if docker_image:
        launch_spec["docker"]["docker_image"] = docker_image

    if "git" not in launch_spec:
        launch_spec["git"] = {}
    if version:
        launch_spec["git"]["version"] = version

    if "overrides" not in launch_spec:
        launch_spec["overrides"] = {}

    if parameters:
        override_args = util._user_args_to_dict(
            launch_spec["overrides"].get("args", [])
        )
        if isinstance(override_args, list):
            base_args = util._user_args_to_dict(
                launch_spec["overrides"].get("args", [])
            )
        elif isinstance(override_args, dict):
            base_args = override_args
        launch_spec["overrides"]["args"] = merge_parameters(parameters, base_args)
    elif isinstance(launch_spec["overrides"].get("args"), list):
        launch_spec["overrides"]["args"] = util._user_args_to_dict(
            launch_spec["overrides"].get("args")
        )
    if entry_point:
        launch_spec["overrides"]["entry_point"] = entry_point

    return launch_spec


def parse_wandb_uri(uri: str) -> Tuple[str, str, str]:
    """Parses wandb uri to retrieve entity, project and run name."""
    uri = uri.split("?")[0]  # remove any possible query params (eg workspace)
    stripped_uri = re.sub(_WANDB_URI_REGEX, "", uri)
    stripped_uri = re.sub(
        _WANDB_DEV_URI_REGEX, "", stripped_uri
    )  # also for testing just run it twice
    stripped_uri = re.sub(
        _WANDB_LOCAL_DEV_URI_REGEX, "", stripped_uri
    )  # also for testing just run it twice
    stripped_uri = re.sub(
        _WANDB_QA_URI_REGEX, "", stripped_uri
    )  # also for testing just run it twice
    entity, project, _, name = stripped_uri.split("/")[1:]
    return entity, project, name


def fetch_wandb_project_run_info(uri: str, api: Api) -> Any:
    """Fetches wandb run info."""
    entity, project, name = parse_wandb_uri(uri)
    try:
        result = api.get_run_info(entity, project, name)
    except CommError as e:
        raise LaunchError(e)
    if result is None:
        raise LaunchError("Run info is invalid or doesn't exist for {}".format(uri))
    if result.get("args") is not None:
        result["args"] = util._user_args_to_dict(result["args"])
    return result


def fetch_project_diff(uri: str, api: Api) -> Optional[str]:
    """Fetches project diff from wandb servers."""
    patch = None
    try:
        entity, project, name = parse_wandb_uri(uri)
        (_, _, patch, _) = api.run_config(project, name, entity)
    except CommError:
        pass
    return patch


def apply_patch(patch_string: str, dst_dir: str) -> None:
    """Applies a patch file to a directory."""
    with open(os.path.join(dst_dir, "diff.patch"), "w") as fp:
        fp.write(patch_string)
    try:
        subprocess.check_call(
            [
                "patch",
                "-s",
                "--directory={}".format(dst_dir),
                "-p1",
                "-i",
                "diff.patch",
            ]
        )
    except subprocess.CalledProcessError:
        raise wandb.Error("Failed to apply diff.patch associated with run.")


def _fetch_git_repo(dst_dir: str, uri: str, version: Optional[str]) -> None:
    """Clones the git repo at ``uri`` into ``dst_dir``.

    checks out commit ``version`` (or defaults to the head commit of the repository's
    master branch if version is unspecified). Assumes authentication parameters are
    specified by the environment, e.g. by a Git credential helper.
    """
    # We defer importing git until the last moment, because the import requires that the git
    # executable is available on the PATH, so we only want to fail if we actually need it.
    import git  # type: ignore

    repo = git.Repo.init(dst_dir)
    origin = repo.create_remote("origin", uri)
    origin.fetch()
    if version is not None:
        try:
            repo.git.checkout(version)
        except git.exc.GitCommandError as e:
            raise ExecutionError(
                "Unable to checkout version '%s' of git repo %s"
                "- please ensure that the version exists in the repo. "
                "Error: %s" % (version, uri, e)
            )
    else:
        repo.create_head("master", origin.refs.master)
        repo.heads.master.checkout()
    repo.submodule_update(init=True, recursive=True)


def merge_parameters(
    higher_priority_params: Dict[str, Any], lower_priority_params: Dict[str, Any]
) -> Dict[str, Any]:
    """Merge the contents of two dicts, keeping values from higher_priority_params if there are conflicts."""
    return {**lower_priority_params, **higher_priority_params}



"""Base SweepConfig classes."""

import yaml
from pathlib import Path

from typing import Union, Dict, List

import jsonschema
from .schema import validator, fill_validate_schema
from copy import deepcopy


def schema_violations_from_proposed_config(config: Dict) -> List[str]:

    schema_violation_messages = []
    for error in validator.iter_errors(config):
        schema_violation_messages.append(f"{error.message}")

    if not isinstance(config, dict):
        raise ValueError("Sweep config must be parsable as a JSON object.")

    # validate min/max - this cannot be done with jsonschema
    # because it does not support comparing values within
    # a json document. so we do it manually here:
    if "parameters" in config:
        for parameter_name, parameter_dict in config["parameters"].items():
            if not isinstance(parameter_dict, Dict):
                raise ValueError(
                    f"Invalid configuration for hyperparameter '{parameter_name}'"
                )
            if "min" in parameter_dict and "max" in parameter_dict:
                # this comparison is type safe because the jsonschema enforces type uniformity
                if parameter_dict["min"] >= parameter_dict["max"]:
                    schema_violation_messages.append(
                        f'{parameter_name}: min {parameter_dict["min"]} is not '
                        f'less than max {parameter_dict["max"]}'
                    )
    return schema_violation_messages


class SweepConfig(dict):
    def __init__(self, d: Dict):
        if not isinstance(d, SweepConfig):
            # ensure the data conform to the schema
            schema_violation_messages = schema_violations_from_proposed_config(d)

            if len(schema_violation_messages) > 0:
                err_msg = "\n".join(schema_violation_messages)
                raise jsonschema.ValidationError(err_msg)

        copied_config = deepcopy(d)
        filled_config = fill_validate_schema(copied_config)
        super(SweepConfig, self).__init__(filled_config)

    def __str__(self) -> str:
        return yaml.safe_dump(self)

    def save(self, filename: Union[Path, str]) -> None:
        with open(filename, "w") as outfile:
            yaml.safe_dump(self, outfile, default_flow_style=False)



from .. import stop_runs, next_run, RunState, SweepRun


def test_hyperband_min_iter_bands():
    sweep_config = {
        "method": "grid",
        "metric": {"name": "loss", "goal": "minimize"},
        "early_terminate": {
            "type": "hyperband",
            "min_iter": 3,
            "eta": 3,
        },
        "parameters": {"a": {"values": [1, 2, 3]}},
    }

    run = next_run(sweep_config, [])
    run.state = RunState.running
    run.history = [{"loss": 10} for _ in range(4)]
    run2 = next_run(sweep_config, [run])
    run2.state = RunState.running
    run2.history = [{"loss": 10 - i} for i in range(10)]
    to_stop = stop_runs(sweep_config, [run, run2])

    assert to_stop[0] is run
    assert to_stop[0].early_terminate_info["bands"][:3] == [3, 9, 27]


def test_hyperband_min_iter_bands_max():
    sweep_config = {
        "method": "grid",
        "metric": {"name": "accuracy", "goal": "maximize"},
        "early_terminate": {
            "type": "hyperband",
            "min_iter": 3,
            "eta": 3,
        },
        "parameters": {"a": {"values": [1, 2, 3]}},
    }

    run = next_run(sweep_config, [])
    run.state = RunState.running
    run.history = [{"accuracy": 10} for _ in range(4)]
    run2 = next_run(sweep_config, [run])
    run2.state = RunState.running
    run2.history = [{"accuracy": 10 + i} for i in range(10)]
    to_stop = stop_runs(sweep_config, [run, run2])

    assert to_stop[0] is run
    assert to_stop[0].early_terminate_info["bands"][:3] == [3, 9, 27]


def test_hyperband_max_iter_bands():
    sweep_config = {
        "method": "grid",
        "metric": {"name": "loss", "goal": "minimize"},
        "early_terminate": {
            "type": "hyperband",
            "max_iter": 81,
            "eta": 3,
            "s": 3,
        },
        "parameters": {"a": {"values": [1, 2, 3]}},
    }

    run = next_run(sweep_config, [])
    run.state = RunState.running
    run.history = [{"loss": 10} for _ in range(4)]
    run2 = next_run(sweep_config, [run])
    run2.state = RunState.running
    run2.history = [{"loss": 10 - i} for i in range(10)]
    to_stop = stop_runs(sweep_config, [run, run2])

    assert to_stop[0] is run
    assert to_stop[0].early_terminate_info["bands"][:3] == [3, 9, 27]


def test_init_from_max_iter():
    sweep_config = {
        "method": "grid",
        "metric": {"name": "loss", "goal": "minimize"},
        "early_terminate": {
            "type": "hyperband",
            "max_iter": 18,
            "eta": 3,
            "s": 2,
        },
        "parameters": {"a": {"values": [1, 2, 3]}},
    }

    run = next_run(sweep_config, [])
    run.state = RunState.running
    run.history = [{"loss": 10} for _ in range(4)]
    run2 = next_run(sweep_config, [run])
    run2.state = RunState.running
    run2.history = [{"loss": 10 - i} for i in range(10)]
    to_stop = stop_runs(sweep_config, [run, run2])

    assert to_stop[0] is run
    assert to_stop[0].early_terminate_info["bands"] == [2, 6]


def test_single_run():

    sweep_config = {
        "method": "grid",
        "metric": {"name": "loss", "goal": "minimize"},
        "early_terminate": {
            "type": "hyperband",
            "max_iter": 18,
            "eta": 3,
            "s": 2,
        },
        "parameters": {"a": {"values": [1, 2, 3]}},
    }

    run = next_run(sweep_config, [])
    run.state = RunState.running
    run.history = [{"loss": 10 - i} for i in range(10)]
    to_stop = stop_runs(sweep_config, [run])
    assert len(to_stop) == 0


def test_2runs_band1_pass():

    sweep_config = {
        "method": "grid",
        "metric": {"name": "loss", "goal": "minimize"},
        "early_terminate": {
            "type": "hyperband",
            "max_iter": 18,
            "eta": 3,
            "s": 2,
        },
        "parameters": {"a": {"values": [1, 2, 3]}},
    }

    run = next_run(sweep_config, [])
    run.state = RunState.running
    run.history = [{"loss": 10}, {"loss": 10}, {"loss": 6}]
    run2 = next_run(sweep_config, [run])
    run2.state = RunState.running
    run2.history = [{"loss": 10 - i} for i in range(10)]
    to_stop = stop_runs(sweep_config, [run, run2])
    assert len(to_stop) == 0


def test_5runs_band1_stop_2():

    sweep_config = {
        "method": "grid",
        "metric": {"name": "loss", "goal": "minimize"},
        "early_terminate": {
            "type": "hyperband",
            "max_iter": 5,
            "eta": 2,
            "s": 2,
        },
        "parameters": {"a": {"values": [1, 2, 3]}},
    }

    runs = [
        SweepRun(
            name="a",
            state=RunState.finished,  # This wont be stopped because already stopped
            history=[
                {"loss": 10},
                {"loss": 9},
            ],
        ),
        SweepRun(
            name="b",
            state=RunState.running,  # This should be stopped
            history=[
                {"loss": 10},
                {"loss": 10},
            ],
        ),
        SweepRun(
            name="c",
            state=RunState.running,  # This passes band 1 but not band 2
            history=[
                {"loss": 10},
                {"loss": 8},
                {"loss": 8},
            ],
        ),
        SweepRun(
            name="d",
            state=RunState.running,
            history=[
                {"loss": 10},
                {"loss": 7},
                {"loss": 7},
            ],
        ),
        SweepRun(
            name="e",
            state=RunState.finished,
            history=[
                {"loss": 10},
                {"loss": 6},
                {"loss": 6},
            ],
        ),
    ]

    to_stop = stop_runs(sweep_config, runs)
    assert to_stop == runs[1:3]


def test_5runs_band1_stop_2_1stnoband():

    sweep_config = {
        "method": "grid",
        "metric": {"name": "loss", "goal": "minimize"},
        "early_terminate": {
            "type": "hyperband",
            "max_iter": 5,
            "eta": 2,
            "s": 2,
        },
        "parameters": {"a": {"values": [1, 2, 3]}},
    }

    runs = [
        SweepRun(
            name="a",
            state=RunState.finished,  # This wont be stopped because already stopped
            history=[
                {"loss": 10},
            ],
        ),
        SweepRun(
            name="b",
            state=RunState.running,  # This should be stopped
            history=[
                {"loss": 10},
                {"loss": 10},
            ],
        ),
        SweepRun(
            name="c",
            state=RunState.running,  # This passes band 1 but not band 2
            history=[
                {"loss": 10},
                {"loss": 8},
                {"loss": 8},
            ],
        ),
        SweepRun(
            name="d",
            state=RunState.running,
            history=[
                {"loss": 10},
                {"loss": 7},
                {"loss": 7},
            ],
        ),
        SweepRun(
            name="e",
            state=RunState.finished,
            history=[
                {"loss": 10},
                {"loss": 6},
                {"loss": 6},
            ],
        ),
    ]

    to_stop = stop_runs(sweep_config, runs)
    assert to_stop == runs[1:3]


def test_eta_3():

    sweep_config = {
        "method": "grid",
        "metric": {"name": "loss", "goal": "minimize"},
        "early_terminate": {
            "type": "hyperband",
            "max_iter": 9,
            "eta": 3,
            "s": 2,
        },
        "parameters": {"a": {"values": [1, 2, 3]}},
    }

    runs = [
        SweepRun(
            name="a",
            state=RunState.finished,  # This wont be stopped because already stopped
            history=[
                {"loss": 10},
                {"loss": 9},
            ],
        ),
        SweepRun(
            name="b",
            state=RunState.running,  # This should be stopped
            history=[
                {"loss": 10},
                {"loss": 10},
            ],
        ),
        SweepRun(
            name="c",
            state=RunState.running,  # This fails the first threeshold but snuck in so we wont kill
            history=[
                {"loss": 10},
                {"loss": 8},
                {"loss": 8},
                {"loss": 3},
            ],
        ),
        SweepRun(
            name="d",
            state=RunState.running,
            history=[
                {"loss": 10},
                {"loss": 7},
                {"loss": 7},
                {"loss": 4},
            ],
        ),
        SweepRun(
            name="e",
            state=RunState.running,  # this passes band 1 but doesn't pass band 2
            history=[
                {"loss": 10},
                {"loss": 6},
                {"loss": 6},
                {"loss": 6},
            ],
        ),
    ]

    # bands are at 1 and 3, thresholds are 7 and 4
    to_stop = stop_runs(sweep_config, runs)
    assert to_stop == [runs[1], runs[-1]]


def test_eta_3_max():

    sweep_config = {
        "method": "grid",
        "metric": {"name": "loss", "goal": "maximize"},
        "early_terminate": {
            "type": "hyperband",
            "max_iter": 9,
            "eta": 3,
            "s": 2,
        },
        "parameters": {"a": {"values": [1, 2, 3]}},
    }

    runs = [
        SweepRun(
            name="a",
            state=RunState.finished,  # This wont be stopped because already stopped
            history=[
                {"loss": -10},
                {"loss": -9},
            ],
        ),
        SweepRun(
            name="b",
            state=RunState.running,  # This should be stopped
            history=[
                {"loss": -10},
                {"loss": -10},
            ],
        ),
        SweepRun(
            name="c",
            state=RunState.running,  # This fails the first threeshold but snuck in so we wont kill
            history=[
                {"loss": -10},
                {"loss": -8},
                {"loss": -8},
                {"loss": -3},
            ],
        ),
        SweepRun(
            name="d",
            state=RunState.running,
            history=[
                {"loss": -10},
                {"loss": -7},
                {"loss": -7},
                {"loss": -4},
            ],
        ),
        SweepRun(
            name="e",
            state=RunState.running,  # this passes band 1 but doesn't pass band 2
            history=[
                {"loss": -10},
                {"loss": -6},
                {"loss": -6},
                {"loss": -6},
            ],
        ),
    ]

    # bands are at 1 and 3, thresholds are 7 and 4
    to_stop = stop_runs(sweep_config, runs)
    assert to_stop == [runs[1], runs[-1]]


def test_hyperband_runs_with_nan_metrics():
    # fixes https://sentry.io/share/issue/e6e002283c0447d6ac4defaf58a5d665/
    runs = [
        SweepRun(
            name="c",
            state=RunState.running,
            history=[
                {"overall_eval_loss": 19.68895149230957},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
                {"overall_eval_loss": "NaN"},
            ],
        ),
        SweepRun(
            name="d",
            state=RunState.finished,
            history=[
                {"overall_eval_loss": 6.214405059814453},
                {"overall_eval_loss": 6.050266265869141},
                {"overall_eval_loss": 5.525775909423828},
                {"overall_eval_loss": 5.143455982208252},
                {"overall_eval_loss": 4.64034366607666},
                {"overall_eval_loss": 4.741260528564453},
                {"overall_eval_loss": 4.611443519592285},
                {"overall_eval_loss": 4.968087673187256},
                {"overall_eval_loss": 4.743710517883301},
                {"overall_eval_loss": 4.415197849273682},
                {"overall_eval_loss": 4.741386413574219},
                {"overall_eval_loss": 5.067233085632324},
                {"overall_eval_loss": 4.5482563972473145},
                {"overall_eval_loss": 4.0536723136901855},
                {"overall_eval_loss": 4.09683895111084},
                {"overall_eval_loss": 4.25504732131958},
                {"overall_eval_loss": 4.438877582550049},
                {"overall_eval_loss": 4.651344299316406},
                {"overall_eval_loss": 3.8646578788757324},
                {"overall_eval_loss": 4.26999568939209},
                {"overall_eval_loss": 4.551799774169922},
                {"overall_eval_loss": 4.867758274078369},
                {"overall_eval_loss": 4.54986572265625},
                {"overall_eval_loss": 4.639401912689209},
                {"overall_eval_loss": 4.570009231567383},
                {"overall_eval_loss": 4.1927008628845215},
                {"overall_eval_loss": 4.218729496002197},
                {"overall_eval_loss": 4.279715538024902},
                {"overall_eval_loss": 4.14383602142334},
                {"overall_eval_loss": 4.373849391937256},
                {"overall_eval_loss": 4.477828025817871},
                {"overall_eval_loss": 4.309680461883545},
                {"overall_eval_loss": 4.336307048797607},
                {"overall_eval_loss": 4.2981953620910645},
                {"overall_eval_loss": 4.259463310241699},
                {"overall_eval_loss": 4.340827465057373},
                {"overall_eval_loss": 4.13071870803833},
                {"overall_eval_loss": 4.145589351654053},
                {"overall_eval_loss": 4.247255802154541},
                {"overall_eval_loss": 4.490720272064209},
                {"overall_eval_loss": 4.3509697914123535},
                {"overall_eval_loss": 4.2368855476379395},
                {"overall_eval_loss": 3.785416841506958},
                {"overall_eval_loss": 4.1248555183410645},
                {"overall_eval_loss": 4.16594123840332},
                {"overall_eval_loss": 4.210123062133789},
                {"overall_eval_loss": 4.134521007537842},
                {"overall_eval_loss": 4.481637954711914},
                {"overall_eval_loss": 3.8986079692840576},
                {"overall_eval_loss": 4.7607035636901855},
                {"overall_eval_loss": 4.429342746734619},
                {"overall_eval_loss": 4.2787957191467285},
                {"overall_eval_loss": 4.423038482666016},
                {"overall_eval_loss": 4.479197978973389},
                {"overall_eval_loss": 4.2208333015441895},
                {"overall_eval_loss": 4.325165748596191},
                {"overall_eval_loss": 4.398067951202393},
                {"overall_eval_loss": 4.34964656829834},
                {"overall_eval_loss": 4.7197394371032715},
                {"overall_eval_loss": 4.056498050689697},
                {"overall_eval_loss": 4.42212438583374},
            ],
        ),
        SweepRun(
            name="e",
            state=RunState.finished,  # this passes band 1 but doesn't pass band 2
            history=[
                {"overall_eval_loss": 6.784717559814453},
                {"overall_eval_loss": 5.467315196990967},
                {"overall_eval_loss": 4.862574577331543},
                {"overall_eval_loss": 5.149237155914307},
                {"overall_eval_loss": 4.713217258453369},
                {"overall_eval_loss": 4.645257472991943},
                {"overall_eval_loss": 4.654730319976807},
                {"overall_eval_loss": 4.218390464782715},
                {"overall_eval_loss": 4.540946960449219},
                {"overall_eval_loss": 4.710428714752197},
                {"overall_eval_loss": 4.327051162719727},
                {"overall_eval_loss": 4.395434379577637},
                {"overall_eval_loss": 4.212771892547607},
                {"overall_eval_loss": 4.131608963012695},
                {"overall_eval_loss": 4.5795159339904785},
                {"overall_eval_loss": 4.4573750495910645},
                {"overall_eval_loss": 4.383418560028076},
                {"overall_eval_loss": 4.822948455810547},
                {"overall_eval_loss": 4.4155707359313965},
                {"overall_eval_loss": 4.176084995269775},
                {"overall_eval_loss": 4.340216636657715},
                {"overall_eval_loss": 4.1993794441223145},
                {"overall_eval_loss": 4.623438835144043},
                {"overall_eval_loss": 4.61974573135376},
                {"overall_eval_loss": 4.306433200836182},
                {"overall_eval_loss": 4.174667835235596},
                {"overall_eval_loss": 4.4847493171691895},
                {"overall_eval_loss": 3.8551318645477295},
                {"overall_eval_loss": 4.469892978668213},
                {"overall_eval_loss": 4.2765727043151855},
                {"overall_eval_loss": 4.316288948059082},
                {"overall_eval_loss": 4.273141384124756},
                {"overall_eval_loss": 4.58651876449585},
                {"overall_eval_loss": 4.162075519561768},
                {"overall_eval_loss": 4.0801591873168945},
                {"overall_eval_loss": 4.499518871307373},
                {"overall_eval_loss": 4.574384689331055},
                {"overall_eval_loss": 4.025749683380127},
                {"overall_eval_loss": 4.137139320373535},
                {"overall_eval_loss": 4.478872299194336},
                {"overall_eval_loss": 4.130046367645264},
                {"overall_eval_loss": 4.551371097564697},
                {"overall_eval_loss": 4.507683753967285},
                {"overall_eval_loss": 4.316148281097412},
                {"overall_eval_loss": 4.449378490447998},
                {"overall_eval_loss": 4.43292236328125},
                {"overall_eval_loss": 3.856088876724243},
                {"overall_eval_loss": 3.8799805641174316},
                {"overall_eval_loss": 4.233527660369873},
                {"overall_eval_loss": 4.096866607666016},
                {"overall_eval_loss": 4.5571112632751465},
                {"overall_eval_loss": 4.188418865203857},
                {"overall_eval_loss": 3.938608407974243},
                {"overall_eval_loss": 4.292693138122559},
                {"overall_eval_loss": 4.055037021636963},
                {"overall_eval_loss": 4.307272434234619},
                {"overall_eval_loss": 4.0329108238220215},
                {"overall_eval_loss": 4.152763366699219},
                {"overall_eval_loss": 4.2551350593566895},
                {"overall_eval_loss": 3.7090814113616943},
                {"overall_eval_loss": 4.495356559753418},
            ],
        ),
    ]
    config = {
        "method": "bayes",
        "metric": {"goal": "minimize", "name": "overall_eval_loss"},
        "parameters": {
            "lr": {"values": [0.001, 0.0003, 0.0001, 3e-05]},
            "pred_out_channels": {"max": 64, "min": 1, "distribution": "int_uniform"},
            "ia_char_embedding_size": {
                "max": 16,
                "min": 1,
                "distribution": "int_uniform",
            },
            "ja_char_embedding_size": {
                "max": 16,
                "min": 1,
                "distribution": "int_uniform",
            },
            "pred_resnet_type_index": {"values": [0, 1, 2]},
            "sr_pred_hidden_channels": {
                "max": 64,
                "min": 1,
                "distribution": "int_uniform",
            },
            "ia_mental_embedding_size": {
                "max": 64,
                "min": 1,
                "distribution": "int_uniform",
            },
            "ja_mental_embedding_size": {
                "max": 64,
                "min": 1,
                "distribution": "int_uniform",
            },
            "goal_pred_hidden_channels": {
                "max": 64,
                "min": 1,
                "distribution": "int_uniform",
            },
            "ia_char_hidden_input_size": {
                "max": 64,
                "min": 1,
                "distribution": "int_uniform",
            },
            "ia_char_resnet_type_index": {"values": [0, 1, 2]},
            "ja_char_hidden_input_size": {
                "max": 64,
                "min": 1,
                "distribution": "int_uniform",
            },
            "ja_char_resnet_type_index": {"values": [0, 1, 2]},
            "ia_char_hidden_output_size": {
                "max": 128,
                "min": 1,
                "distribution": "int_uniform",
            },
            "ja_char_hidden_output_size": {
                "max": 128,
                "min": 1,
                "distribution": "int_uniform",
            },
            "action_pred_hidden_channels": {
                "max": 64,
                "min": 1,
                "distribution": "int_uniform",
            },
            "ia_mental_hidden_input_size": {
                "max": 64,
                "min": 1,
                "distribution": "int_uniform",
            },
            "ia_mental_resnet_type_index": {"values": [0, 1, 2]},
            "ja_mental_hidden_input_size": {
                "max": 64,
                "min": 1,
                "distribution": "int_uniform",
            },
            "ja_mental_resnet_type_index": {"values": [0, 1, 2]},
            "ia_mental_hidden_output_size": {
                "max": 64,
                "min": 1,
                "distribution": "int_uniform",
            },
            "ja_mental_hidden_output_size": {
                "max": 64,
                "min": 1,
                "distribution": "int_uniform",
            },
            "pred_hidden_channel_type_index": {
                "values": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
            },
            "ia_char_hidden_channel_type_index": {
                "values": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
            },
            "ja_char_hidden_channel_type_index": {
                "values": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
            },
            "ia_mental_hidden_channel_type_index": {
                "values": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
            },
            "ja_mental_hidden_channel_type_index": {
                "values": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
            },
        },
        "early_terminate": {"s": 2, "eta": 3, "type": "hyperband", "max_iter": 300},
    }
    assert stop_runs(config, runs) == []
